

import warnings
warnings.simplefilter('ignore')

import unittest

class TestCreate(unittest.TestCase):

    def x_test_basic(self):

        from metapack.support import metapack_defaults_file
        from metapack import MetapackDoc

        d = MetapackDoc(metapack_defaults_file())

        d['Root'].set_terms(
            Title='This is the title',
            Description='Description!',
            Origin='busboom.org',
            Dataset='foobar')

        d['Contacts'].set_terms(
            Wrangler=('Eric Busboom', {'Email':'eric@busboom.org', 'Url':'http://busboom.org'})
        )

        d.sort_by_term()
        d.update_name()

        print(d.as_lines())



if __name__ == '__main__':
    unittest.main()



import logging
from metapack.constants import PACKAGE_PREFIX

from tabulate import tabulate

logger = logging.getLogger('user')
logger_err = logging.getLogger('cli-errors')
debug_logger = logging.getLogger('debug')


def cli_init(log_level=logging.INFO):
    import sys

    from metapack.appurl import SearchUrl

    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter('%(message)s'))
    out_hdlr.setLevel(log_level)
    logger.addHandler(out_hdlr)
    logger.setLevel(log_level)

    out_hdlr = logging.StreamHandler(sys.stderr)
    out_hdlr.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    out_hdlr.setLevel(logging.WARN)
    logger_err.addHandler(out_hdlr)
    logger_err.setLevel(logging.WARN)

    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter('DEBUG: %(message)s'))
    out_hdlr.setLevel(log_level)
    debug_logger.addHandler(out_hdlr)
    debug_logger.setLevel(log_level)

    SearchUrl.initialize()  # Setup the JSON index search.


def prt(*args, **kwargs):
    logger.info(' '.join(str(e) for e in args), **kwargs)


def warn(*args, **kwargs):
    logger_err.warning(' '.join(str(e) for e in args), **kwargs)


def err(*args, **kwargs):
    import sys
    logger_err.critical(' '.join(str(e) for e in args), **kwargs)
    sys.exit(1)


def metatab_info(cache):
    pass


def new_metatab_file(mt_file, template):
    from os.path import exists
    from uuid import uuid4

    from metatab.util import make_metatab_file

    template = template if template else 'metatab'

    if not exists(mt_file):
        doc = make_metatab_file(template)

        doc['Root']['Identifier'] = str(uuid4())

        doc.write_csv(mt_file)


def find_files(base_path, types):
    from os import walk
    from os.path import join, splitext

    for root, dirs, files in walk(base_path):
        if '_metapack' in root:
            continue

        for f in files:
            if f.startswith('_'):
                continue

            b, ext = splitext(f)
            if ext[1:] in types:
                yield join(root, f)


def get_lib_module_dict(doc):
    return doc.get_lib_module_dict()


def dump_resources(doc):
    for r in doc.resources():
        prt(r.name, r.resolved_url)


def dump_resource(doc, name, lines=None):
    import unicodecsv as csv
    import sys
    from itertools import islice
    from tabulate import tabulate
    from rowgenerators.rowpipe.exceptions import CasterExceptionError, TooManyCastingErrors

    r = doc.resource(name=name)

    if not r:
        err("Did not get resource for name '{}'".format(name))

    # WARNING! This code will not generate errors if line is set ( as for the -H
    # option because the errors are tansfered from the row pipe to the resource after the
    # iterator is exhausted

    gen = islice(r, 1, lines)

    def dump_errors(error_set):
        for col, errors in error_set.items():
            warn("Errors in casting column '{}' in resource '{}' ".format(col, r.name))
            for error in errors:
                warn("    ", error)

    try:
        if lines and lines <= 20:
            try:
                prt(tabulate(list(gen), list(r.headers)))
            except TooManyCastingErrors as e:
                dump_errors(e.errors)
                err(e)

        else:

            w = csv.writer(sys.stdout.buffer)

            if r.headers:
                w.writerow(r.headers)
            else:
                warn("No headers for resource '{}'; have schemas been generated? ".format(name))

            for row in gen:
                w.writerow(row)

    except CasterExceptionError as e:  # Really bad errors, not just casting problems.
        raise e
        err(e)
    except TooManyCastingErrors as e:
        dump_errors(e.errors)
        err(e)

    dump_errors(r.errors)


def dump_schema(doc, name):
    from tabulate import tabulate

    t = get_table(doc, name)

    rows = []
    header = 'name altname datatype description'.split()
    for c in t.children:
        cp = c.properties
        rows.append([cp.get(h) for h in header])

    prt(tabulate(rows, header))


def get_table(doc, name):
    t = doc.find_first('Root.Table', value=name)

    if not t:

        table_names = ["'" + t.value + "'" for t in doc.find('Root.Table')]

        if not table_names:
            table_names = ["<No Tables>"]

        err("Did not find schema for table name '{}' Tables are: {}"
            .format(name, " ".join(table_names)))

    return t



def _exec_build(p, package_root, force,  nv_name, extant_url_f, post_f):
    from metapack import MetapackUrl

    if force:
        reason = 'Forcing build'
        should_build = True
    elif p.is_older_than_metadata():
        reason = 'Metadata was changed after package'
        should_build = True
    elif not p.exists():
        reason = "Package doesn't exist"
        should_build = True
    else:
        reason = 'Metadata was changed before package'
        should_build = False

    if should_build:
        prt("Building {} package ({})".format(p.type_code, reason))
        url = p.save()
        prt("Package ( type: {} ) saved to: {}".format(p.type_code, url))
        created = True
    else:
        prt("Not building {} package ({})".format(p.type_code, reason))

    if not should_build and p.exists():
        created = False
        url = extant_url_f(p)

    post_f()

    if nv_name:
        p.move_to_nv_name()

    return p, MetapackUrl(url, downloader=package_root.downloader), created


def make_excel_package(file, package_root, cache, env, force, nv_name=None, nv_link=False):
    from metapack.package import ExcelPackageBuilder

    assert package_root

    p = ExcelPackageBuilder(file, package_root, callback=prt, env=env)

    return _exec_build(p,
                       package_root, force,  nv_name,
                       lambda p: p.package_path.path,
                       lambda: p.create_nv_link() if nv_link else None)


def make_zip_package(file, package_root, cache, env, force, nv_name=None, nv_link=False):
    from metapack.package import ZipPackageBuilder

    assert package_root

    p = ZipPackageBuilder(file, package_root, callback=prt, env=env)

    return _exec_build(p,
                       package_root, force, nv_name,
                       lambda p: p.package_path.path,
                       lambda: p.create_nv_link() if nv_link else None)


def make_filesystem_package(file, package_root, cache, env, force, nv_name=None, nv_link=False):
    from os.path import join

    from metapack.package import FileSystemPackageBuilder
    from metatab import DEFAULT_METATAB_FILE

    assert package_root

    p = FileSystemPackageBuilder(file, package_root, callback=prt, env=env)

    return _exec_build(p,
                       package_root, force, nv_name,
                       lambda p: join(p.package_path.path.rstrip('/'), DEFAULT_METATAB_FILE),
                       lambda: p.create_nv_link() if nv_link else None )


def make_csv_package(file, package_root, cache, env, force, nv_name=None, nv_link=False):
    from metapack.package import CsvPackageBuilder

    assert package_root

    p = CsvPackageBuilder(file, package_root, callback=prt, env=env)

    return _exec_build(p,
                       package_root, force,  nv_name,
                       lambda p: p.package_path.path,
                       lambda: p.create_nv_link() if nv_link else None)


def make_s3_package(file, package_root, cache, env, skip_if_exists, acl='public-read'):
    from metapack import MetapackUrl
    from metapack.package import S3PackageBuilder

    assert package_root

    p = S3PackageBuilder(file, package_root, callback=prt, env=env, acl=acl)

    if not p.exists() or not skip_if_exists:
        url = p.save()
        prt("Packaged saved to: {}".format(url))
        created = True
    elif p.exists():
        prt("S3 Filesystem Package already exists")
        created = False
        url = p.access_url

    return p, MetapackUrl(url, downloader=file.downloader), created


def update_name(mt_file, fail_on_missing=False, report_unchanged=True, force=False):
    from metapack import MetapackDoc

    if isinstance(mt_file, MetapackDoc):
        doc = mt_file
    else:
        doc = MetapackDoc(mt_file)

    o_name = doc.find_first_value("Root.Name", section=['Identity', 'Root'])

    updates = doc.update_name(force=force, report_unchanged=report_unchanged)

    for u in updates:
        prt(u)

    prt("Name is: ", doc.find_first_value("Root.Name", section=['Identity', 'Root']))

    if o_name != doc.find_first_value("Root.Name", section=['Identity', 'Root']) or force:
        write_doc(doc, mt_file)


def write_doc(doc, mt_file):
    """
    Write a Metatab doc to a CSV file, and update the Modified time
    :param doc:
    :param mt_file:
    :return:
    """

    from rowgenerators import parse_app_url

    import subprocess

    try:
        out = subprocess.run(['git', 'remote', 'show', 'origin'], stdout=subprocess.PIPE,
                             stderr=subprocess.DEVNULL, timeout=6) \
            .stdout.decode('utf-8')

        fetchline = next(l.split() for l in out.splitlines() if 'Fetch' in l)
    except (TimeoutError, StopIteration, subprocess.TimeoutExpired, FileNotFoundError):
        fetchline = None

    if fetchline:
        t = doc['Root'].get_or_new_term('GitUrl')
        t.value = fetchline[-1]

    u = parse_app_url(mt_file)

    if u.scheme == 'file':
        doc.write_csv(mt_file)
        return True
    else:
        return False
        # warn("Not writing back to url ", mt_file)


def process_schemas(mt_file, cache=None, clean=False, report_found=True):
    from rowgenerators.exceptions import SourceError, SchemaError
    from requests.exceptions import ConnectionError
    from itertools import islice

    from metapack import MetapackDoc
    from rowgenerators.source import SelectiveRowGenerator
    from tableintuit import TypeIntuiter

    if isinstance(mt_file, MetapackDoc):
        doc = mt_file
        write_doc_to_file = False
    else:
        doc = MetapackDoc(mt_file)
        write_doc_to_file = True

    try:
        if clean:
            doc['Schema'].clean()
        else:
            doc['Schema']

    except KeyError:
        doc.new_section('Schema', ['DataType', 'Altname', 'Description'])

    schemas_processed = 0

    for r in doc['Resources'].find('Root.Resource'):

        schema_term = r.schema_term

        col_count = len(list(r.columns()))
        datatype_count = sum(1 for c in r.columns() if c['datatype'])

        if schema_term and col_count == datatype_count:
            if report_found:
                prt("Found table for '{}'; skipping".format(r.schema_name))
            continue

        if col_count != datatype_count:

            prt("Found table for '{}'; but {} columns don't have datatypes"
                .format(r.schema_name, col_count-datatype_count))

        prt("Processing {}".format(r.name))
        schemas_processed += 1


        try:
            slice = islice(r.row_generator, 500)
        except SchemaError:
            warn("Failed to build row processor table, using raw row generator")
            slice = islice(r.raw_row_generator, 500)

        si = SelectiveRowGenerator(slice,
                                   headers=[int(i) for i in r.get_value('headerlines', '0').split(',')],
                                   start=int(r.get_value('startline', 1)))

        try:
            ti = TypeIntuiter().run(si)
        except SourceError as e:
            warn("Failed to process resource '{}'; {}".format(r.name, e))
            continue
        except ConnectionError as e:
            warn("Failed to download resource '{}'; {}".format(r.name, e))
            continue

        if schema_term:
            table = schema_term
            prt("Updating table '{}' ".format(r.schema_name))

            col_map = {c.name.lower(): c for c in table.children if c.term_is('Table.Column')}

            for i, c in enumerate(ti.to_rows()):
                extant = col_map.get(c['header'])
                if extant:
                    extant.datatype = type_map.get(c['resolved_type'], c['resolved_type'])

        else:

            table = doc['Schema'].new_term('Table', r.schema_name)
            prt("Adding table '{}' ".format(r.schema_name))

            for i, c in enumerate(ti.to_rows()):
                raw_alt_name = alt_col_name(c['header'], i)
                alt_name = raw_alt_name if raw_alt_name != c['header'] else ''

                t = table.new_child('Column', c['header'],
                                    datatype=type_map.get(c['resolved_type'], c['resolved_type']),
                                    altname=alt_name)

    if write_doc_to_file and schemas_processed:
        write_doc(doc, mt_file)


def extract_path_name(ref):
    from os.path import basename
    from os.path import splitext, abspath

    from rowgenerators import parse_app_url

    uparts = parse_app_url(ref)

    ss = parse_app_url(ref)

    if not uparts.scheme:
        path = abspath(ref)
        name = basename(splitext(path)[0])
    else:
        path = ref

        v = ss.target_file if ss.target_file else uparts.path

        name = basename(splitext(v)[0])

    return path, name


def alt_col_name(name, i):
    import re

    if not name:
        return 'col{}'.format(i)

    return re.sub('_+', '_', re.sub('[^\w_]', '_', str(name)).lower()).rstrip('_')


type_map = {
    float.__name__: 'number',
    int.__name__: 'integer',
    bytes.__name__: 'string',
    str.__name__: 'text',

    # Numpy types
    'datetime64[ns]': 'datetime'
}


class MetapackCliMemo(object):
    def __init__(self, args, downloader):
        from os import getcwd
        from os.path import join

        from metatab import DEFAULT_METATAB_FILE

        self.cwd = getcwd()
        self.args = args

        self.downloader = downloader

        self.cache = self.downloader.cache

        self.config = get_config()

        frag = ''

        if isinstance(args.metatabfile, (list, tuple)):
            args.metatabfile = args.metatabfile.pop(0)

        # Just the fragment was provided
        if args.metatabfile and args.metatabfile.startswith('#'):
            frag = args.metatabfile
            mtf = None
        else:
            frag = ''
            mtf = args.metatabfile

        # If could not get it from the args, Set it to the default file name in the current dir
        if not mtf:
            mtf = join(self.cwd, DEFAULT_METATAB_FILE)

        self.init_stage2(mtf, frag)

    def init_stage2(self, mtf, frag):
        from metapack import MetapackUrl
        from rowgenerators import parse_app_url

        self.frag = frag
        self.mtfile_arg = mtf + frag

        u = parse_app_url(self.mtfile_arg, downloader=self.downloader)

        if u.scheme == 'index':
            self.mtfile_url = MetapackUrl(u.resolve(), downloader=self.downloader)
        else:
            self.mtfile_url = MetapackUrl(self.mtfile_arg, downloader=self.downloader)

        self.resource = self.mtfile_url.resource_name

        # This is probably a bug in AppUrl, but I'm not sure.
        if self.mtfile_url.resource_format == 'zip' and self.resource == 'metadata.csv':
            self.resource = None


        self.package_url = self.mtfile_url.package_url
        self.mt_file = self.mtfile_url.metadata_url

        #assert self.package_url.scheme == 'file'

        if hasattr(self.args, 'build_directory') and self.args.build_directory:
            self.package_root = parse_app_url(self.args.build_directory)
        else:
            self.package_root = self.package_url.join(PACKAGE_PREFIX)

        assert self.package_root._downloader

    @property
    def doc(self):
        from metapack import MetapackDoc
        return MetapackDoc(self.mt_file)


def get_config():
    """Return a configuration dict"""
    from os import environ
    from os.path import expanduser
    from pathlib import Path
    import yaml

    def pexp(p):
        try:
            return Path(p).expanduser()
        except AttributeError:
            # python 3.4
            return Path(expanduser(p))

    paths = [environ.get("METAPACK_CONFIG"), '~/.metapack.yaml', '/etc/metapack.yaml']

    for p in paths:
        try:
            if pexp(p).exists():
                with  pexp(p).open() as f:
                    config = yaml.safe_load(f)
                    config['_loaded_from'] = str(pexp(p))
                    return config
        except TypeError:
            pass

    return {}



def update_index(packages, package_path, suffix=''):
    from os import listdir
    from os.path import join, exists, isdir, splitext

    raise DeprecationWarning()

    if packages:
        # Just update one packages
        add_package_to_index(package_path, packages, suffix=suffix)

    else:
        # Build the whole package index
        packages = []

        def yield_packages(d):

            for e in listdir(d):
                path = join(d, e)
                bn, ext = splitext(path)
                if isdir(path):
                    if exists(join(path, 'metadata.csv')):
                        yield join(path, 'metadata.csv')
                elif ext in ('.xls', '.xlsx', '.zip'):
                    yield path

        for p in yield_packages(package_path):
            add_package_to_index(p, packages, suffix=None)

    return packages



def new_search_index():
    return []


def list_rr(doc):

    d = []
    for r in doc.resources():
        d.append(('Resource', r.name, r.url))

    for r in doc.references():
        d.append(('Reference', r.name, r.url))


    prt(tabulate(d, 'Type Name Url'.split()))


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

""" """


import json
import shutil
from genericpath import exists, getmtime
from os import getcwd, makedirs, remove, getenv
from os.path import join, dirname, isdir

from nbconvert.writers import FilesWriter
from rowgenerators import parse_app_url
from metatab.datapackage import convert_to_datapackage
from metatab import DEFAULT_METATAB_FILE
from .core import PackageBuilder
from metapack.util import ensure_dir, write_csv, slugify, datetime_now
from metapack.appurl import MetapackUrl, SearchUrl

def count_decl(doc):
    decls = doc.find('Root.Declare')
    assert len(doc.terms) == 0 or len(decls) == 1, (len(decls), len(doc.terms) )



class FileSystemPackageBuilder(PackageBuilder):
    """Build a filesystem package"""

    type_code = 'fs'
    type_suffix = ''

    def __init__(self, source_ref, package_root, callback=None, env=None):

        super().__init__(source_ref, package_root,  callback, env)

        if not self.package_root.isdir():
            self.package_root.ensure_dir()

        self.package_path, self.cache_path = self.make_package_path(self.package_root, self.package_name)

        self.doc_file = self.package_path.join(DEFAULT_METATAB_FILE)

    @classmethod
    def make_package_path(cls, package_root, package_name):

        cache_path = join(package_name, DEFAULT_METATAB_FILE)

        package_path = package_root.join(package_name)

        return package_path, cache_path

    def exists(self):

        try:
            path = self.doc_file.path
        except AttributeError:
            path = self.doc_file

        return self.package_path.isdir() and exists(path)

    def remove(self):

        if self.package_path.is_dir():
            shutil.rmtree(self.package_path.path)

    def is_older_than_metadata(self):
        """
        Return True if the package save file is older than the metadata. If it is, it should be rebuilt.  Returns
        False if the time of either can't be determined

        :param path: Optional extra save path, used in save_path()

        """

        try:
            path = self.doc_file.path
        except AttributeError:
            path = self.doc_file

        source_ref = self._doc.ref.path

        try:
            age_diff = getmtime(source_ref) - getmtime(path)

            return age_diff > 0

        except (FileNotFoundError, OSError):
            return False

    def save(self):

        self.check_is_ready()

        self.sections.resources.sort_by_term()

        self.doc.cleanse()

        self.load_declares()

        self._load_documentation_files()

        self._load_resources()

        self._load_files()

        self._write_dpj()

        self._clean_doc()

        doc_file = self._write_doc()

        self._write_html()

        return doc_file

    def _write_doc(self):

        self._doc['Root'].get_or_new_term('Root.Issued').value = datetime_now()

        self._doc.write_csv(self.doc_file)
        return MetapackUrl(self.doc_file, downloader=self._downloader)

    def _write_dpj(self):

        with open(join(self.package_path.path, 'datapackage.json'), 'w') as f:
            f.write(json.dumps(convert_to_datapackage(self._doc), indent=4))

    def _write_html(self):

        with open(join(self.package_path.path, 'index.html'), 'w') as f:
            f.write(self._doc.html)


    def _load_resource(self, source_r, abs_path=False):
        """The CSV package has no resources, so we just need to resolve the URLs to them. Usually, the
            CSV package is built from a file system ackage on a publically acessible server. """

        from itertools import islice
        from metapack.exc import MetapackError

        # Refetch the resource ... IIRC b/c the source_r resource may actuall be from
        # a different package. So r is the resource we want to possibly modify in this package,
        # while source_r is from a different souce package, whose data is being loaded into this
        # one.
        r = self.datafile(source_r.name)

        self.prt("Loading data for '{}' ".format(r.name))

        if not r.name:
            raise MetapackError(f"Resource/reference term has no name: {str(r)}")

        if r.term_is('root.sql'):
            new_r = self.doc['Resources'].new_term('Root.Datafile', '')
            new_r.name = r.name

            self.doc.remove_term(r)

            r = new_r

        r.url = 'data/' + r.name + '.csv'

        path = join(self.package_path.path, r.url)

        makedirs(dirname(path), exist_ok=True)

        if exists(path):
            remove(path)

        gen = islice(source_r, 1, None)
        headers = source_r.headers
        self.write_csv(path, headers, gen)

        for k, v in source_r.post_iter_meta.items():
            r[k] = v

        # Writing between resources so row-generating programs and notebooks can
        # access previously created resources. We have to clean the doc before writing it

        ref = self._write_doc()

        # What a wreck ... we also have to get rid of the 'Transform' values, since the CSV files
        # that are written don't need them, and a lot of intermediate processsing ( specifically,
        # jupyter Notebooks, ) does not load them.
        p = FileSystemPackageBuilder(ref, self.package_root)
        p._clean_doc()

        ref = p._write_doc()

    def _load_documentation_files(self):

        from metapack.jupyter.exporters import DocumentationExporter

        notebook_docs = []

        # First find and remove notebooks from the docs. These wil ge processed to create
        # normal documents.
        try:
            for term in list(self.doc['Documentation'].find('Root.Documentation')):
                u = parse_app_url(term.value)
                if u.target_format == 'ipynb':
                    notebook_docs.append(term)
                    self.doc.remove_term(term)
        except KeyError:
            self.warn("No documentation defined in metadata")


        # Process all of the normal files
        super()._load_documentation_files()

        de = DocumentationExporter()
        fw = FilesWriter()
        fw.build_directory = join(self.package_path.path,'docs')

        # Now, generate the documents directly into the filesystem package
        for term in notebook_docs:
            u = parse_app_url(term.value)
            nb_path = u.path(self.source_dir)

            output, resources = de.from_filename(nb_path)
            fw.write(output, resources, notebook_name='notebook')

            de.update_metatab(self.doc, resources)

    def _load_documentation(self, term, contents, file_name):

        try:
            title = term['title'].value
        except KeyError:
            self.warn("Documentation has no title, skipping: '{}' ".format(term.value))
            return

        path = join(self.package_path.path,  file_name)

        self.prt("Loading documentation for '{}', '{}' ".format(title, file_name))

        makedirs(dirname(path), exist_ok=True)

        if exists(path):
            remove(path)

        with open(path, 'wb') as f:
            f.write(contents)

    def _load_file(self,  filename, contents):

        if "__pycache__" in filename:
            return

        path = join(self.package_path.path, filename)

        ensure_dir(dirname(path))

        with open(path, 'wb') as f:
            f.write(contents)


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
CLI program for managing packages
"""

from metapack import Downloader
from metapack.jupyter.convert import convert_documentation, convert_notebook, convert_wordpress,\
    extract_metatab, convert_hugo
from metapack.cli.core import prt, err, warn, get_config
from os import environ
from os.path import basename
import re

downloader = Downloader()


def notebook(subparsers):
    parser = subparsers.add_parser(
        'notebook',
        help='Convert Metatab-formatted Jupyter notebooks. ',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    parser.set_defaults(run_command=run_notebook)

    parser.add_argument('notebook', nargs='?',
                        help="Path to a notebok file' ")

    parser.add_argument('-p', '--profile', help="Name of a BOTO or AWS credentails profile", required=False)

    parser.add_argument('--exceptions', default=False, action='store_true',
                        help='Show full stack tract for some unhandled exceptions')

    parser.set_defaults(handler=None)

    ##
    ## Build Group

    build_group = parser.add_argument_group('Building Metatab Files', 'Build and manage a metatab file for a pacakge')

    build_group.add_argument('-m', '--metatab', default=False, action='store_true',
                             help='Extract the metatab file')

    build_group.add_argument('-l', '--lines', default=False, action='store_true',
                             help='When displaying a Metatab file, display as lines, not as CSV')

    build_group.add_argument('-P', '--package', default=False, action='store_true',
                             help='Build a package from a Jupyter notebook')

    build_group.add_argument('-D', '--documentation', default=False, action='store_true',
                             help='With -M, make only the documentation')

    build_group.add_argument('-H', '--hugo', default=False, nargs='?',
                             help='Write images and Markdown into a Hugo static site directory. or use '
                                  'METAPACK_HUGO_DIR env var')

    build_group.add_argument('-w', '--wordpress', default=False, nargs='?',
                             help='Publish the notebook to a Wordpress blog as a post. Arg refers to credentials in the config')

    build_group.add_argument('-W', '--wordpress-page', default=False, nargs='?',
                             help='Publish the notebook to a Wordpress blog as a page. Arg refers to credentials in the config')

def run_notebook(args):

    # Maybe need to convert a notebook first
    if args.package:
        convert_notebook(args.notebook)

    elif args.documentation:
        convert_documentation(args.notebook)

    elif args.metatab:
        doc = extract_metatab(args.notebook)

        if args.lines:
            for line in doc.lines:
                print(": ".join(line))
        else:
            print(doc.as_csv())

    elif args.hugo:
        convert_hugo(args.notebook,args.hugo)

    elif args.wordpress:

        from os import makedirs
        from os.path import exists
        from metapack.util import ensure_dir

        p = '/tmp/metapack-wp-notebook/'
        ensure_dir(p)

        output_file, resources = convert_wordpress(args.notebook, p)

        r, post = publish_wp(args.wordpress, output_file, resources)
        prt("Post url: ", post.link)

def get_site_config(site_name):
    from textwrap import dedent
    config = get_config()

    if not config:
        err("No metatab configuration found. Can't get Wordpress credentials")

    site_config = config.get('wordpress', {}).get(site_name, {})

    if not site_config:
        err("In config file '{}', expected 'wordpress.{}' section for site config"
            .format(config['_loaded_from'], site_name))

    if 'url' not in site_config or 'user' not in site_config or 'password' not in site_config:
        err(dedent(
            """
            Incomplete configuration. Expected:
                wordpress.{site_name}.url
                wordpress.{site_name}.user
                wordpress.{site_name}.password
            In configuration file '{cfg_file}'
            """.format(site_name=site_name, cfg_file=config['_loaded_from'])
        ))

    return site_config['url'], site_config['user'], site_config['password']

def prepare_image(slug, file_name, post_id):
    from xmlrpc import client as xmlrpc_client

    # prepare metadata
    data = {
        'name': 'picture.jpg',
        'type': 'image/jpeg',  # mimetype
    }

    # read the binary file and let the XMLRPC library encode it into base64
    with open(file_name, 'rb') as img:
        return {
            'name': '{}-{}'.format(slug,basename(file_name)),
            'type': 'image/png',  # mimetype
            'bits': xmlrpc_client.Binary(img.read()),
            'post_id': post_id
        }


def publish_wp(site_name, output_file, resources):
    import json
    from uuid import uuid4
    from wordpress_xmlrpc import Client, WordPressPost
    from wordpress_xmlrpc.methods.posts import GetPosts, NewPost, EditPost, GetPost
    from wordpress_xmlrpc.methods.users import GetUserInfo
    from wordpress_xmlrpc.methods.media import UploadFile, GetMediaLibrary

    # http://busboom.org/wptest/wp-content/uploads/sites/7/2017/11/output_16_0-300x200.png

    url, user, password = get_site_config(site_name)

    meta = {}
    for r in resources:
        if r.endswith('.json'):
            with open(r) as f:
                meta = json.load(f)

    fm = meta.get('frontmatter',{})

    if not 'identifier' in fm or not fm['identifier']:
        err("Can't publish notebook without a unique identifier. Add this to the "
            "Metatab document or frontmatter metadata:\n   identifier: {}".format(str(uuid4())))

    wp = Client(url, user, password)

    def cust_field_dict(post):
        return dict( (e['key'],e['value']) for e in post.custom_fields)

    post = None
    for _post in wp.call(GetPosts()):
        if cust_field_dict(_post).get('identifier') == fm['identifier']:
            post = _post
            prt("Updating old post")
            break

    if not post:
        post = WordPressPost()
        post.id = wp.call(NewPost(post))
        prt("Creating new post")

    post.title = fm.get('title','')
    post.slug = fm.get('slug')

    with open(output_file) as f:
        content = f.read()

    post.terms_names = {
        'post_tag': ['test', 'firstpost'],
        'category': ['Introductions', 'Tests']
    }
    post.custom_fields = [
        {'key': 'identifier', 'value': fm['identifier']},

    ]

    post.excerpt = meta.get('bref', meta.get('description'))

    def strip_image_name(n):
        """Strip off the version number from the media file"""
        from os.path import splitext
        import re
        return re.sub(r'\-\d+$','',splitext(n)[0])


    extant_files = list(wp.call(GetMediaLibrary(dict(parent_id=post.id))))

    def find_extant_image(image_name):
        for img in extant_files:
            if strip_image_name(basename(img.metadata['file'])) == strip_image_name(image_name):
                return img

        return None

    for r in resources:

        image_data = prepare_image(fm['identifier'], r, post.id)
        img_from = "/{}/{}".format(fm['slug'], basename(r))

        extant_image = find_extant_image(image_data['name'])

        if  extant_image:
            prt("Post already has image:", extant_image.link)
            img_to = extant_image.link

        elif r.endswith('.png'): # Foolishly assuming all images are PNGs
            response = wp.call(UploadFile(image_data, overwrite=True))

            prt("Uploaded image {} to id={}, {}".format(basename(r), response['id'], response['link']))

            img_to = response['link']

        content = content.replace(img_from, img_to)


    post.content = content

    r = wp.call(EditPost(post.id, post))

    return r,  wp.call(GetPost(post.id))


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Metapack CLI program for creating new metapack package directories
"""

from metapack.package import *

from .core import MetapackCliMemo as _MetapackCliMemo, err, prt, warn

downloader = Downloader()


class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)

        self.term = self.args.term[0]
        self.value = self.args.value[0] if hasattr(self.args, 'value') else None

        parts = self.term.split('.')

        if len(parts) != 2 and len(parts) != 3:
            err('Term arg must have 2 or 3 words seperated by a period')

        if len(parts) == 3:
            self.section, parts = parts[0], parts[1:]
            self.term = '.'.join(parts)
        else:
            self.section = 'Root'


def edit_args(subparsers):
    parser = subparsers.add_parser(
        'edit',
        help='Edit a metatab file'
    )

    cmds = parser.add_subparsers(help='Commands')

    parser.set_defaults(run_command=edit_cmd)

    cp_add = cmds.add_parser('add', help='Add a term')
    cp_add.set_defaults(edit_command=add_cmd)
    cp_add.add_argument('term', nargs=1,
                        help="A fully qualified term. Prefix with a third component to set the section ( 'Resources.Root.Datafile' ) ")
    cp_add.add_argument('value', nargs=1, help="The term value, for add and edit ")
    cp_add.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    cp_edit = cmds.add_parser('change', help='Edit a term')
    cp_edit.set_defaults(edit_command=change_cmd)
    cp_edit.add_argument('term', nargs=1,
                         help="A fully qualified term. Prefix with a third component to set the section ( 'Resources.Root.Datafile' ) ")
    cp_edit.add_argument('value', nargs=1, help="The term value, for add and edit ")
    cp_edit.add_argument('metatabfile', nargs='?',
                         help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    cp_delete = cmds.add_parser('delete', help='Delete a term')
    cp_delete.set_defaults(edit_command=delete_cmd)
    cp_delete.add_argument('term', nargs=1,
                           help="A fully qualified term. Prefix with a third component to set the section ( 'Resources.Root.Datafile' ) ")
    cp_delete.add_argument('metatabfile', nargs='?',
                           help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

def edit_cmd(args):
    m = MetapackCliMemo(args, downloader)
    m.args.edit_command(m)


def add_cmd(m):
    print('ADD', m.term, m.value, m.doc.ref)

    doc = m.doc

    term = doc[m.section].new_term(m.term, m.value)

    print(doc.as_csv())
    doc.write_csv()

def change_cmd(m):

    doc = m.doc

    t = doc.find_first(m.term, section=m.section)

    t.value = m.value

    print(doc.as_csv())
    doc.write_csv()

def delete_cmd(m):

    doc = m.doc

    terms = doc.find(m.term, section=m.section)

    if not terms:
        warn("No terms found for: '{}' in section '{}' ".format(m.term, m.section))

    for term in terms:
        doc.remove_term(term)

    terms = doc.find(m.term, section=m.section)

    if terms:
        warn("Delete failed")

    print(doc.as_csv())
    doc.write_csv()





def test_data(*paths):
    from os.path import dirname, join, abspath
    import metapack.test.test_data

    return join(dirname(abspath(metapack.test.test_data.__file__)), *paths)


def cache_fs():

    from fs.tempfs import TempFS

    return TempFS('rowgenerator')

def get_cache():
    return cache_fs()


import unittest


class TestIssues(unittest.TestCase):
    """Test Metapack AppUrls and Row Generators"""

    def test_ref_resource_confusion(self):
        from metapack import open_package
        u = 'http://library.metatab.org/ffiec.gov-cra_disclosure_smb_orig-2010_2015-2.csv'

        p = open_package(u)

        self.assertEqual(1, len(list(r for r in p['Resources'].find('Root.Resource'))))


    def test_no_reference_descriptions(self):

        from metapack import MetapackDoc
        from metapack.test.support import test_data
        from metatab.generate import TextRowGenerator

        mt_lines="""
Declare: metatab-latest
Identifier: foobar
Name: sandiegodata.org-cra_demo-1

Dataset: foobar
Origin: civiknowledge.com
Version: 1


Section: References

Reference: metapack+http://library.metatab.org/ffiec.gov-cra_aggregate_smb-orig-4.csv#sb_agg_loan_orig
Reference.Name: agg_loan_orig
Reference.Description: CRA Loan originations, aggregated to tracts.
        """

        doc = MetapackDoc(TextRowGenerator(mt_lines))

        r = doc.reference('agg_loan_orig')

        for c in r.columns():
            print(c.get('name'), c.get('description'))

    def test_refs_not_using_schemas(self):

        from metapack import  MetapackDoc
        from rowgenerators.valuetype import ShapeValue
        from metapack.test.support import test_data
        from metatab.generate import TextRowGenerator

        with open(test_data('line', 'line-oriented-doc.txt')) as f:
            text = f.read()

        doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest\n" + text))

        r = doc.reference('sra_geo')

        self.assertEqual('civicknowledge.com-rcfe_affordability-2015-4',r.doc.name)
        self.assertEqual('sangis.org-census_regions-2010-sandiego-7', r.resolved_url.resource.doc.name)

        rs = r.resolved_url.resource

        self.assertIsNotNone(rs.row_processor_table())

        self.assertEqual([int,int,str,ShapeValue],  list(type(c) for c in list(rs)[1]))

        self.assertIsNone(r.row_processor_table())

        self.assertEqual([int, int, str, ShapeValue], list(type(c) for c in list(r)[1]))

        for c in rs.columns():
            print(c.get('description'))

        print('---')

        for c in r.columns():
            print(c.get('description'))


    def x_test_csv_join(self):


        from rowgenerators import parse_app_url
        import metapack as mp

        pkg = mp.open_package('http://library.metatab.org/noaa.gov-localclimate-200808_201807-san-3.csv')

        for t in pkg['Documentation'].find('Root.Documentation'):
            u = parse_app_url(t.value)

            t = pkg.package_url.join_target(u).get_resource().get_target()
            print("u=",u)
            print("pu=",pkg.package_url)
            print("jt=",pkg.package_url.join_target(u))
            print("R=",pkg.package_url.join_target(u).get_resource() )
            print('A', t._fragment)


    def x_test_wack_doc_urls(self):
        """Getting inline documentation fails when the package URL is for an online CSV package"""

        import metapack as mp

        from rowgenerators import parse_app_url

        pkg = mp.open_package('http://library.metatab.org/noaa.gov-localclimate-200808_201807-san-3.csv')

        for t in pkg['Documentation'].find('Root.IncludeDocumentation'):
            u = parse_app_url(t.value)

            t = pkg.package_url.join_target(u).get_resource().get_target()
            print("u=",u)
            print("pu=",pkg.package_url)
            print("jt=",pkg.package_url.join_target(u))
            print("R=",pkg.package_url.join_target(u).get_resource() )
            print('A', t._fragment)


        pkg = mp.open_package('http://library.metatab.org/noaa.gov-localclimate-200808_201807-san-3.zip')

        for t in pkg['Documentation'].find('Root.IncludeDocumentation'):
            u = parse_app_url(t.value)

            t = pkg.package_url.join_target(u).get_resource().get_target()
            self.assertTrue(t.exists())
            self.assertTrue(t.path.endswith('README.md'))

        pkg = mp.open_package('/Users/eric/Library/Application '
                              'Support/metapack/library.metatab.org/noaa.gov-localclimate-200808_201807-san-3/')

        for t in pkg['Documentation'].find('Root.IncludeDocumentation'):
            u = parse_app_url(t.value)
            t = pkg.package_url.join_target(u).get_resource().get_target()
            self.assertTrue(t.exists())
            self.assertTrue(t.path.endswith('README.md'))

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Functions for executing Jupyter notebooks
"""

from .exc import NotebookError


def execute_notebook(nb_path, pkg_dir, dataframes, write_notebook=False, env=None):
    """
    Execute a notebook after adding the prolog and epilog. Can also add %mt_materialize magics to
    write dataframes to files

    :param nb_path: path to a notebook.
    :param pkg_dir: Directory to which dataframes are materialized
    :param dataframes: List of names of dataframes to materialize
    :return: a Notebook object
    """

    import nbformat
    from metapack.jupyter.preprocessors import AddEpilog, AddProlog
    from metapack.jupyter.exporters import ExecutePreprocessor, Config
    from os.path import dirname, join, splitext, basename
    from nbconvert.preprocessors.execute import CellExecutionError

    with open(nb_path) as f:
        nb = nbformat.read(f, as_version=4)

    root, ext = splitext(basename(nb_path))

    c = Config()

    nb, resources = AddProlog(config=c, env=env or {}).preprocess(nb, {})

    nb, resources = AddEpilog(config=c, pkg_dir=pkg_dir,
                              dataframes=dataframes,
                              ).preprocess(nb, {})

    try:
        ep = ExecutePreprocessor(config=c)

        nb, _ = ep.preprocess(nb, {'metadata': {'path': dirname(nb_path)}})
    except (CellExecutionError, TimeoutError) as e:
        err_nb_path = join(dirname(nb_path), root + '-errors' + ext)
        with open(err_nb_path, 'wt') as f:
            nbformat.write(nb, f)

        raise CellExecutionError("Errors executing noteboook. See notebook at {} for details.\n{}"
                                 .format(err_nb_path, ''))

    except ImportError as e:
        raise NotebookError("Failed to import a library required for notebook execution: {}".format(str(e)))


    if write_notebook:
        if write_notebook is True:
            exec_nb_path = join(dirname(nb_path), root + '-executed' + ext)
        else:
            exec_nb_path = write_notebook

        with open(exec_nb_path, 'wt') as f:
            nbformat.write(nb, f)

    return nb


# -*- coding: utf-8 -*
# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

jupyter nbextension install --py metatab.jupyter.magic

"""

from __future__ import print_function

import sys

import logging
import os
from IPython import get_ipython
from IPython.core.magic import Magics, magics_class, line_magic, cell_magic, line_cell_magic
from IPython.core.magic_arguments import (argument, magic_arguments,
                                          parse_argstring)
from IPython.display import display, HTML, Latex
from metapack import MetapackDoc
from metapack.appurl import MetapackPackageUrl
from metapack.cli.core import process_schemas
from metapack.html import bibliography, data_sources
from metatab import TermParser
from metatab.generate import TextRowGenerator
from os import makedirs, getcwd
from os.path import join, abspath, dirname, exists, normpath
from rowgenerators import Downloader, parse_app_url
from warnings import warn

logger = logging.getLogger('user')
logger_err = logging.getLogger('cli-errors')
doc_logger = logging.getLogger('doc')
debug_logger = logging.getLogger('debug_logger')

MT_DOC_VAR = 'mt_pkg' # Namespace name for the metatab document.

@magics_class
class MetatabMagic(Magics):

    """Magics for using Metatab in Jupyter Notebooks

    """



    @property
    def mt_doc(self):
        """Return the current metatab document, which must be created with either %%metatab
        or %mt_load_package"""



        if MT_DOC_VAR not in self.shell.user_ns:

            package_url = MetapackPackageUrl("metapack+file:" + os.getcwd() + '/', downloader=Downloader())

            self.shell.user_ns[MT_DOC_VAR] = \
                MetapackDoc(TextRowGenerator("Declare: metatab-latest\n"), package_url=package_url)

            inline_doc = self.shell.user_ns[MT_DOC_VAR]

            if not 'Resources' in inline_doc:
                inline_doc.new_section('Resources', ['Name', 'Description'])
            if not 'Resources' in inline_doc:
                inline_doc.new_section('References', ['Name', 'Description'])

            # Give all of the sections their standard args, to make the CSV versions of the doc
            # prettier
            for name, s in inline_doc.sections.items():
                try:
                    s.args = inline_doc.decl_sections[name.lower()]['args']
                except KeyError:
                    pass

        return self.shell.user_ns[MT_DOC_VAR]

    @property
    def package_dir(self):
        """Return the current metatab document, which must be created with either %%metatab
        or %mt_load_package"""

        return self.shell.user_ns['_package_dir']

    def add_term_lines(self, text):


        assert 'root.reference' in TermParser.term_classes

        tp = TermParser(TextRowGenerator(text), resolver= self.mt_doc.resolver, doc= self.mt_doc)

        self.mt_doc.load_terms(tp)


    def clean_doc(self, doc):

        # Some sections have terms that are unique on Name

        for sec in ['Resources','References','Bibliography', 'Schema', 'Contacts']:
            try:
                seen = set()
                for t in doc[sec]:
                    if t.name in seen:
                        doc.remove_term(t)
                    seen.add(t.name)
            except KeyError:
                pass

        # For the Root, just make every term unique name term name, keeping the last one

        seen = {}
        for t in list(doc['Root']):
            if t.term in seen:
                extant = seen[t.term]
                doc.remove_term(extant)
            seen[t.term] = t

    @magic_arguments()
    @argument('-s', '--show', help='After loading, display the document', action='store_true')
    @argument('-p', '--package_dir', help='Set the directory where the package will be created')
    @cell_magic
    def metatab(self, line, cell):
        """Process a cell of Metatab data, in line format. Stores document in the `mt_pkg` variable
        """

        args = parse_argstring(self.metatab, line)

        inline_doc = self.mt_doc

        self.add_term_lines(cell)

        extant_identifier = inline_doc.get_value('Root.Identifier')
        extant_name = inline_doc.get_value('Root.Name')

        inline_doc.update_name(force=True, create_term=True)

        process_schemas(inline_doc)

        self.clean_doc(inline_doc)

        if args.show:
            for l in inline_doc.lines:
                print(': '.join(str(e) for e in l))

        if args.package_dir:
            self.shell.user_ns['_package_dir'] = abspath(join(getcwd(), args.package_dir))
        else:
            self.shell.user_ns['_package_dir'] = join(getcwd(), inline_doc.get_value('Root.Name'))

        if extant_identifier != inline_doc.get_value('Root.Identifier'):
            print("Identifier updated.  Set 'Identifier: {}'  in document".format(inline_doc.get_value('Root.Identifier')))

        if extant_name != inline_doc.get_value('Root.Name'):
            print("Name Changed. Set 'Name: {}'  in document".format(inline_doc.get_value('Root.Name')))


    @magic_arguments()
    @argument('-s', '--source', help='Force opening the source package', action='store_true')
    @line_magic
    def mt_open_package(self, line):
        """Find the metatab file for this package, open it, and load it into the namespace. """

        from metapack.jupyter.ipython import open_package

        args = parse_argstring(self.mt_open_package, line)
        self.shell.user_ns[MT_DOC_VAR] = open_package(self.shell.user_ns)


        if self.mt_doc.package_url:
            u = parse_app_url(self.mt_doc.package_url)

    @line_magic
    def mt_import_terms(self, line):
        """Import the value of some Metatab terms into the notebook namespace """

        mt_terms = {}

        doc = self.mt_doc

        mt_terms['root.title'] = doc['Root'].find_first_value('Root.Title')
        mt_terms['root.description'] = doc['Root'].find_first_value('Root.Description')
        mt_terms['root.name'] = doc['Root'].find_first_value('Root.Name')
        mt_terms['root.identifier'] = doc['Root'].find_first_value('Root.Identifier')

        mt_terms['contacts'] = []

        for t in doc.get_section('Contacts', []):
            d = t.as_dict()
            d['type'] = t.record_term_lc
            mt_terms['contacts'].append(d)

        mt_terms['bibliography'] = []

        for t in doc.get_section('Bibliography', []):
            d = t.as_dict()
            d['type'] = t.record_term_lc
            mt_terms['bibliography'].append(d)

        mt_terms['references'] = []

        for t in doc.get_section('References', []):
            d = t.as_dict()
            d['type'] = t.record_term_lc
            mt_terms['references'].append(d)

        print('mt_terms')

        self.shell.user_ns['mt_terms'] = mt_terms

    @line_magic
    def mt_process_schemas(self, line):
        """Add Schema entries for resources to the metatab file. Does not write the doc file"""

        process_schemas(self.mt_doc)

    @magic_arguments()
    @argument('package_dir', help='Package directory')
    @argument('--feather', help='Use feather for serialization', action='store_true' )
    @line_magic
    def mt_materialize_all(self, line):
        """Materialize all of the dataframes that has been previously added with mt_add_dataframe
        """

        from json import dumps

        from metapack.rowgenerator import PandasDataframeSource
        import csv

        materialized = []

        args = parse_argstring(self.mt_materialize_all, line)

        try:
            cache = self.mt_doc._cache
        except KeyError:
            cache = Downloader().cache

        for df_name, ref in self.shell.user_ns.get('_material_dataframes',{}).items():

            u = parse_app_url(args.package_dir).join(ref)

            path = u.path

            if not exists(dirname(path)):
                makedirs(dirname(path))

            df = self.shell.user_ns[df_name].fillna('')

            if args.feather:
                path = path.replace('.csv','.feather')
                df.to_feather(path)
            else:
                gen = PandasDataframeSource(u, df, cache=cache)
                with open(path, 'w') as f:
                    w = csv.writer(f)
                    w.writerows(gen)

            materialized.append({
                'df_name': df_name,
                'path': path,
                'ref':ref
            })

        # Show information about the dataframes that were materialized, so it can be harvested later
        print(dumps(materialized, indent=4))

    @magic_arguments()
    @argument('df_name', help='Dataframe name')
    @argument('dir', help='Path to output directory. Created if it does not exist')
    @line_magic
    def mt_materialize(self, line):
        """Materialize a single dataframe
        """

        from json import dumps
        from rowgenerators import get_cache
        from metapack.rowgenerator import PandasDataframeSource
        from metapack.util import ensure_dir
        import csv
        from os.path import join

        args = parse_argstring(self.mt_materialize, line)

        dr = args.dir.strip("'")

        ensure_dir(dr)

        try:
            cache = self.mt_doc._cache
        except KeyError:
            cache = get_cache()

        path = join(dr, args.df_name + ".csv")
        df = self.shell.user_ns[args.df_name].fillna('')
        gen = PandasDataframeSource(parse_app_url(path), df, cache=cache)

        with open(path, 'w') as f:
            w = csv.writer(f)
            w.writerows(gen)

        print(dumps({
            'df_name': args.df_name,
            'path': path,

        }, indent=4))


    @line_magic
    def mt_show_metatab(self, line):
        """Dump the metatab file to the output, so it can be harvested after the notebook is executed"""

        try:
            for line in self.mt_doc.lines:

                if  line[1]: # Don't display "None"
                    print(': '.join(line))
        except KeyError:
            pass

    @magic_arguments()
    @argument('-m','--materialize', help='Save the data for the dataframe during package conversion', action="store_true")
    @argument('-n','--name', help='Metadata reference name of the dataframe')
    @argument('-t','--title', help='Title of the dataframe')
    @argument('-d', '--dump', help='Dump example schema for the dataframe. Implied when used as line magic', action="store_true")
    @argument('--feather', help='Materialize with feather', action="store_true")
    @argument('dataframe_name', nargs=1, help='Variable name of the dataframe name')
    @line_cell_magic
    def mt_add_dataframe(self, line, cell=''):
        """Add a dataframe to a metatab document's data files

        """
        from metapack.jupyter.core import process_schema
        from metatab.exc import ParserError

        args = parse_argstring(self.mt_add_dataframe, line)

        if not cell:
            is_line = True
            args.dump = True
        else:
            is_line = False

        dataframe_name = args.dataframe_name[0]

        if not '_material_dataframes' in self.shell.user_ns:
            self.shell.user_ns['_material_dataframes'] = {}

        df = self.shell.user_ns[dataframe_name]

        try:
            cell_doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest\n" + cell))
        except ParserError as e:
            warn('Failed to parse Metatab in cell: {} '.format(e))
            return

        cell_table = cell_doc.find_first('Root.Table')

        if cell_table and ( args.name or args.title ):
            warn("The name and title arguments are ignored when the cell includes a Metatab table definition")

        if cell_table:
            name = cell_table.get_value('name')
            title = cell_table.get_value('title','')
            description = cell_table.get_value('description','')
        else:
            name = None
            title = ''
            description = ''


        if not name:
            name = args.name or dataframe_name;

        if not title:
            title = args.title or dataframe_name

        if not name:
            warn("Name must be set with .name property, or --name option".format(args['<dataframe_name>']))
            return

        title = title.strip("'").strip('"')

        try:
            doc = self.mt_doc
        except KeyError:
            doc = None

        if args.materialize:
            ref = 'file:data/{}.csv'.format(name)
            self.shell.user_ns['_material_dataframes'][dataframe_name] = ref

        elif doc is not None:
            ref = 'ipynb:notebooks/{}.ipynb#{}'.format( doc.as_version(None), dataframe_name)

        else:
            ref = None

        table = None

        resource_term = None

        ##
        ## First, process the schema, extracting the columns from the dataframe.
        ##

        if doc and ref:
            if not 'Resources' in doc:
                doc.new_section('Resources')

            resource_term = doc['Resources'].get_or_new_term("Root.Datafile", ref)

            resource_term['name'] = name
            resource_term['title'] = title
            resource_term['description'] = description

            df = df.reset_index()

            table = process_schema(doc, doc.resource(name), df)

            if not table:
                table = doc['Schema'].find_first('Root.Table', name)

        ##
        ## Next, apply the names from  table description from the cell
        ##

        if cell_table:

            cols_by_name = {c.name:c for c in cell_table.find('Table.Column') }

            for i,c in enumerate(table.find('Table.Column')):

                cell_column = cols_by_name.get(c.name)
                try:
                    cell_col_by_pos = list(cols_by_name.values())[i]
                except KeyError:
                    cell_col_by_pos = None
                except IndexError:
                    cell_col_by_pos = None

                if cell_column:
                    c.description = cell_column.description
                    c.name = cell_column.name
                elif cell_col_by_pos:
                    c.description = cell_col_by_pos.description
                    c.name = cell_col_by_pos.name


        if args.dump and table:
            print("Table:", resource_term.name)

            if resource_term and resource_term.title:
                print("Table.Title:", resource_term.get_value('title'))
                print("Table.Description:", resource_term.get_value('description') if resource_term.get_value('description')  else '')

            for c in table.find('Table.Column'):
                print("Table.Column:", c.name)
                print("  .Datatype:", c.datatype)
                print("  .Description:",  c.description or '')

            if is_line:
                print("\nCopy the above into the cell, and change to a cell magic, with '%%' ")


    @magic_arguments()
    @argument('--format', help="Format, html or latex. Defaults to 'all' ", default='all', nargs='?', )
    @argument('converters', help="Class names for citation converters", nargs='*', )
    @line_magic
    def mt_bibliography(self, line):
        """Display, as HTML, the bibliography for the metatab document. With no argument,
         concatenate all doc, or with an arg, for only one. """

        args = parse_argstring(self.mt_bibliography, line)

        def import_converter(name):
            components = name.split('.')
            mod = __import__(components[0])
            for comp in components[1:]:
                mod = getattr(mod, comp)
            return mod

        converters = [ import_converter(e) for e in args.converters ]

        if args.format == 'html' or args.format == 'all':
            display(HTML(bibliography(self.mt_doc, converters=converters, format='html')))

        if args.format == 'latex' or args.format == 'all':
            display(Latex(bibliography(self.mt_doc, converters=converters, format='latex')))

    @magic_arguments()
    @argument('--format', help="Format, html or latex. Defaults to 'all' ", default='all', nargs='?', )
    @argument('converters', help="Class names for citation converters", nargs='*', )
    @line_magic
    def mt_data_references(self, line):
        """Display, as HTML, the bibliography for the metatab document. With no argument,
         concatenate all doc, or with an arg, for only one. """

        args = parse_argstring(self.mt_bibliography, line)

        def import_converter(name):
            components = name.split('.')
            mod = __import__(components[0])
            for comp in components[1:]:
                mod = getattr(mod, comp)
            return mod

        converters = [import_converter(e) for e in args.converters]

        if args.format == 'html' or args.format == 'all':
            display(HTML(data_sources(self.mt_doc, converters=converters, format='html')))

        if args.format == 'latex' or args.format == 'all':
            display(Latex(data_sources(self.mt_doc, converters=converters, format='latex')))

    @magic_arguments()
    @argument('lib_dir', help='Directory', nargs='?',)
    @line_magic
    def mt_lib_dir(self, line):
        """Declare a source code directory and add it to the sys path

        The argument may be a directory, a URL to a Metatab ZIP archive, or a reference to a
        Root.Reference or Root.Resource term that references a Metatab ZIP Archive

        If lib_dir is not specified, it defaults to 'lib'

        If lib_dir is a directory, the target is either at the same level as the CWD or
        one level up.

        If lib_dir is a URL, it must point to a Metatab ZIP archive that has an interal Python
        package directory. The URL may hav path elements after the ZIP archive to point
        into the ZIP archive. For instance:

            %mt_lib_dir http://s3.amazonaws.com/library.metatab.org/ipums.org-income_homevalue-5.zip

        If lib_dir is anything else, it is a reference to the name of a Root.Reference or Root.Resource term that
        references a Metatab ZIP Archive. For instance:

            %%metatab
            ...
            Section: References
            Reference: metatab+http://s3.amazonaws.com/library.metatab.org/ipums.org-income_homevalue-5.zip#income_homeval
            ...


            %mt_lib_dir incv
            from lib.incomedist import *

        """

        from os.path import splitext, basename

        args = parse_argstring(self.mt_lib_dir, line) # Its a normal string

        if not args.lib_dir:
            lib_dir = 'lib'

        else:
            lib_dir = args.lib_dir

        if not '_lib_dirs' in self.shell.user_ns:
            self.shell.user_ns['_lib_dirs'] = set()

        u = parse_app_url(lib_dir)

        # Assume files are actually directories
        # This clause will pickup both directories and reference names, but ref names should not
        # exist as directories.
        if u.proto == 'file':

            lib_dir = normpath(lib_dir).lstrip('./')

            for path in [ abspath(lib_dir), abspath(join('..',lib_dir))]:
                if exists(path) and path not in sys.path:
                    sys.path.insert(0,path)
                    self.shell.user_ns['_lib_dirs'].add(lib_dir)
                    return

        # Assume URLS are to Metapack packages on the net
        if (u.scheme == 'https' or u.scheme == 'http'):

            # The path has to be a Metatab ZIP archive, and the root directory must be the same as
            # the name of the path

            r = u.get_resource()

            pkg_name, _ = splitext(basename(r.path))

            lib_path = r.join(pkg_name).path

            if lib_path not in sys.path:
                sys.path.insert(0,lib_path)

        # Assume anything else is a Metatab Reference term name
        elif self.mt_doc.find_first('Root.Reference', name=lib_dir) or self.mt_doc.resource(lib_dir) :

            r = self.mt_doc.find_first('Root.Reference', name=lib_dir) or self.mt_doc.resource(lib_dir)

            ur = parse_app_url(r.url).inner

            return self.mt_lib_dir(str(ur))

        else:
            logger.error("Can't find library directory: '{}' ".format(lib_dir))


    @line_magic
    def mt_show_libdirs(self, line):
        """Dump the list of lib dirs as JSON"""
        import json

        if '_lib_dirs' in self.shell.user_ns:
            print(json.dumps(list(self.shell.user_ns['_lib_dirs'])))
        else:
            print(json.dumps([]))



@magics_class
class MetapackMagic(Magics):

    """Magics for using Metatab in Jupyter Notebooks

    """

    @magic_arguments()
    @argument('dataframe_name',  help='Variable name of the dataframe name')
    @argument('--doc', help='Variable name of the document. Defaults to the local source package')
    @argument('--description', help='Description of the resource')
    @argument('-r','--reset-index', action='store_true', help='Reset the dataframe index before adding')
    @line_magic
    def mp_add_dataframe(self, line, cell=''):

        from metapack.util import get_materialized_data_cache
        from .ipython import add_dataframe, open_source_package
        from json import dumps
        from metapack.rowgenerator import PandasDataframeSource
        import csv
        from os.path import join

        args = parse_argstring(self.mp_add_dataframe, line)

        if args.doc:
            doc = self.shell.user_ns[args.doc]
        else:
            doc = open_source_package()

        if self.shell.user_ns.get("METAPACK_BUILDING"): # var set by AddProlog
            warn("Building, so materializing instead of adding to document")

            cache = get_materialized_data_cache(doc)

            path = join(cache, args.dataframe_name + ".csv")
            df = self.shell.user_ns[args.dataframe_name].fillna('')
            gen = PandasDataframeSource(parse_app_url(path), df, cache=cache)

            with open(path, 'w') as f:
                w = csv.writer(f)
                w.writerows(gen)

            print(dumps({
                'df_name': args.dataframe_name,
                'path': path,

            }, indent=4))

        else:
            # We're in interactive mode, so just records the dataset in the metadata
            df = self.shell.user_ns[args.dataframe_name]

            add_dataframe(df if not args.reset_index else df.reset_index(),
                          args.dataframe_name, pkg=doc, description=args.description.strip("'"))


def load_ipython_extension(ipython):
    # In order to actually use these magics, you must register them with a
    # running IPython.  This code must be placed in a file that is loaded once
    # IPython is up and running:
    ip = get_ipython()
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.
    ip.register_magics(MetatabMagic)
    ip.register_magics(MetapackMagic)

    # init_logging()


def unload_ipython_extension(ipython):
    # If you want your extension to be unloadable, put that logic here.
    pass



# -*- coding: utf-8 -*
# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""


from metapack.exc import MetapackError

class NotebookError(MetapackError):
    pass



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""

from metapack.cli.core import prt, err
from metapack.package import *
from .core import MetapackCliMemo as _MetapackCliMemo
from tabulate import tabulate
import sys
from pkg_resources import get_distribution, DistributionNotFound, iter_entry_points

downloader = Downloader()

class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)
        self


def info_args(subparsers):

    parser = subparsers.add_parser(
        'info',
        help='Print info about a package '
    )

    parser.set_defaults(run_command=info)

    group = parser.add_mutually_exclusive_group()

    group.add_argument('-n', '--name', default=False, action='store_true',
                             help="Print the name, with version")
    group.add_argument('-N', '--root-name', default=False, action='store_true',
                       help="Print the name, without the version")

    group.add_argument('-r', '--resources', default=False, action='store_true',
                       help="List the resources in the package")

    group.add_argument('-s', '--schema', default=False, action='store_true',
                       help="Print a table of the common schema for all resources, or if the metatab file ref has a resource, only that one")

    group.add_argument('-T', '--row-table', default=False, action='store_true',
                       help="Print the row-processor table, including transforms and valuetypes")

    parser.add_argument('-v', '--version', default=False, action='store_true',
                             help='Print Metapack versions')

    parser.add_argument('-V', '--versions', default=False, action='store_true',
                        help='Print version of several important packages')

    parser.add_argument('-C', '--cache', default=False, action='store_true',
                        help='Print the location of the cache')

    parser.add_argument('-M', '--materialized', default=False, action='store_true',
                        help='Print the location of the materialized data cache')

    parser.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

def info(args):
    from metapack.exc import MetatabFileNotFound

    m = MetapackCliMemo(args, downloader)


    try:
        if m.args.name:
            prt(m.doc.name)

        elif m.args.resources:
            list_rr(m.doc)

        elif m.args.root_name:
            prt(m.doc.as_version(None))

        elif m.args.schema:
            dump_schemas(m)

        elif m.args.row_table:
            dump_rptable(m)

        elif args.version:
            prt(get_distribution('metapack'))

        elif args.cache:
            from shlex import quote
            prt(downloader.cache.getsyspath('/'))

        elif args.materialized:
            from metapack.util import get_materialized_data_cache
            prt(get_materialized_data_cache())

        elif args.versions:
            print_versions(m)

        else:
            try:
                prt(m.doc.name)
            except MetatabFileNotFound:
                print_versions(m)

    except MetatabFileNotFound:
        err('No metatab file found')





def print_versions(m):
    from pkg_resources import EntryPoint
    from tabulate import tabulate

    main_packages = ('metapack', 'metatab', 'metatabdecl', 'rowgenerators', 'publicdata', 'tableintuit')

    packages = []
    for pkg_name in main_packages:
        try:
            d = get_distribution(pkg_name)
            packages.append([d.project_name, d.version])

        except (DistributionNotFound, ModuleNotFoundError) as e:
            # package is not installed

            pass

    prt(tabulate(packages, headers='Package Version'.split()))
    prt('')
    prt(tabulate([(ep.name, ep.dist) for ep in iter_entry_points(group='mt.subcommands')],
                                                                headers='Subcommand Package Version'.split()))


def list_rr(doc):

    d = []
    for r in doc.resources():
        d.append(('Resource', r.name, r.url))

    prt(tabulate(d, 'Type Name Url'.split()))

def get_resource(m):

    r = m.doc.resource(m.resource)

    if not r:
        prt('\nSelect a resource to display the schema for:\n')
        list_rr(m.doc)
        prt('')
        prt("Specify the resource as a fragment, escaping it if the '#' is the first character. For instance: ")
        prt("  mp info -s \#resource_name")
        prt('')
        sys.exit(0)

    return r

def dump_schemas(m):

    r = get_resource(m)

    st = r.schema_term
    rows_about_columns = []
    for c in st.find('Table.Column'):
        rows_about_columns.append((c.name, c.get_value('altname'), c.get_value('datatype'), c.get_value('description')))

    prt(tabulate(rows_about_columns, headers='Name AltName DataType Description'.split()))


def dump_rptable(m):

    r = get_resource(m)

    print(r.row_processor_table())



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""

import mimetypes
import traceback
from os import getenv
from os.path import join, basename

from metapack import MetapackDoc, Downloader, open_package
from metapack.cli.core import err
from metatab import  DEFAULT_METATAB_FILE, MetatabError
from metapack.package.s3 import S3Bucket
from .core import MetapackCliMemo as _MetapackCliMemo
from .core import prt, warn, write_doc


downloader = Downloader()

class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)

        self.api_key = self.args.api or getenv('METAKAN_API_KEY')

        self.ckan_url = self.args.ckan or getenv('METAKAN_CKAN_URL')

    def set_mt_arg(self, metatabfile):


        if not self.ckan_url:
            err("Set the --ckan option or the METAKAN_CKAN_URL env var to set the URL of a ckan instance")

        if not self.api_key:
            err("Set the --api option METAKAN_API_KEY env var  with the API key to a CKAN instance")

    def update_mt_arg(self, metatabfile):
        """Return a new memo with a new metatabfile argument"""
        o = MetapackCliMemo(self.args)
        o.set_mt_arg(metatabfile)
        return o

def metakan(subparsers):


    parser = subparsers.add_parser(
        'ckan',
        help='CKAN management of Metatab packages'
    )

    parser.set_defaults(run_command=run_ckan)

    parser.add_argument('-i', '--info', default=False, action='store_true',
                   help="Show configuration information")

    parser.add_argument('-c', '--ckan', help="URL for CKAN instance")

    parser.add_argument('-a', '--api', help="CKAN API Key")

    parser.add_argument('-p', '--packages', action='store_true',
                        help="The file argument is a text file with a list of package URLs to load")

    parser.add_argument('-C', '--configure', action='store_true',
                        help="Load groups and organizations from a metatab file")

    parser.add_argument('-D', '--dump', action='store_true',
                        help="Dump groups and organizations to a metatab file")

    parser.add_argument('metatabfile', nargs='?', default=DEFAULT_METATAB_FILE,
                        help='Path to a Metatab file, or an s3 link to a bucket with Metatab files. ')


def run_ckan(args):

    m = MetapackCliMemo(args, downloader=downloader)

    if m.mtfile_url.scheme == 's3':
        # Find all of the top level CSV files in a bucket and use them to create CKan entries

        b = S3Bucket(m.mtfile_arg)

        for e in b.list():
            key = e['Key']
            if '/' not in key and key.endswith('.csv'):
                url = b.access_url(key)
                prt("Processing", url)
                send_to_ckan(m.update_mt_arg(url))

    elif m.args.packages:
        # Load a list of packages
        with open(m.mtfile_arg) as f:
            for line in f.readlines():
                url = line.strip()
                prt("Processing", url)
                try:
                    send_to_ckan(m.update_mt_arg(url))
                except Exception as e:
                    warn("Failed to process {}: {}".format(line, e))

    elif m.args.configure:
        configure_ckan(m)

    elif m.args.dump:
        dump_ckan(m)

    else:
        send_to_ckan(m)

    exit(0)

def send_to_ckan(m):

    from ckanapi import RemoteCKAN, NotFound
    try:
        doc = MetapackDoc(m.mt_file, cache=m.cache)

    except (IOError, MetatabError) as e:
        err("Failed to open metatab '{}': {}".format(m.mt_file, e))

    c = RemoteCKAN(m.ckan_url, apikey=m.api_key)

    ckanid = doc.find_first_value('Root.Ckanid')

    unversioned_name = doc.as_version(None)
    ckan_name = unversioned_name.replace('.','-')

    id_name = ckanid or ckan_name

    try:
        pkg = c.action.package_show(name_or_id=id_name)
        prt("Updating CKAN dataset for '{}'".format(id_name))
        found = True
    except NotFound as e:
        e.__traceback__ = None
        traceback.clear_frames(e.__traceback__)
        found = False

    if not found:
        try:
            pkg = c.action.package_show(name_or_id=ckan_name)
            prt("Updating CKAN dataset for '{}'".format(id_name))
            found = True
        except NotFound as e:
            e.__traceback__ = None
            traceback.clear_frames(e.__traceback__)
            found = False

    if not found:
        try:
            pkg = c.action.package_create(name=ckan_name)
        except Exception as e:
            err("Failed to create package for name '{}': {} ".format(ckan_name, e))

        prt("Adding CKAN dataset for '{}'".format(ckan_name))

    pkg['title'] = doc.find_first_value('Root.Title')

    if not pkg['title']:
        pkg['title'] = doc.find_first_value('Root.Description')

    pkg['version'] =  doc.find_first_value('Root.Version')

    pkg['groups'] = [ {'name': g.value } for g in doc['Root'].find('Root.Group')]

    pkg['tags'] = [{'name': g.value} for g in doc['Root'].find('Root.Tag')]

    org_name = doc.get_value('Root.Origin', doc.get_value('Root.CkanOrg'))

    if org_name:
        org_name_slug = org_name.replace('.','-')
        try:

            owner_org = c.action.organization_show(id=org_name_slug).get('id')
            pkg['owner_org'] = owner_org
        except NotFound:
            warn("Didn't find org for '{}'; not setting organization ".format(org_name_slug))
            org_name_slug = None
    else:
        org_name_slug = None

    extras = {}

    for t in doc.find('*.*', section='Root'):
        if not t.term_is('Root.Distribution'):
            extras[t.qualified_term] = t.value

    pkg['extras'] = [ {'key':k, 'value':v} for k, v in extras.items() ]

    resources = []

    # Try to set the markdown from a CSV package, since it will have the
    # correct links.
    markdown = None

    for dist in doc.find('Root.Distribution'):

        prt("Processing {} package: {}".format(dist.type, dist.value))

        package_url = dist.package_url
        metadata_url = dist.metadata_url

        if dist.type == 'zip':
            d = dict(
                url=str(package_url.inner),
                name=basename(package_url.path),
                format='ZIP',
                mimetype=mimetypes.guess_type(package_url.path)[0],
                description='ZIP version of package'
            )
            resources.append(d)
            prt("Adding ZIP package ", d['name'])

        elif dist.type == 'xlsx':
            d = dict(
                url=str(package_url.inner),
                name=basename(package_url.path),
                format='XLSX',
                mimetype=mimetypes.guess_type(package_url.path)[0],
                description='Excel version of package'
            )
            resources.append(d)
            prt("Adding XLS package ", d['name'])

        elif dist.type == 'csv':

            d=dict(
                url=str(package_url.inner),
                name=basename(package_url.path),
                format='csv',
                mimetype=mimetypes.guess_type(metadata_url.path)[0],
                description='CSV Package Metadata in Metatab format'
            )

            resources.append(d)
            prt("Adding {} package {}".format(d['format'], d['name']))

            try:
                p = open_package(metadata_url)
            except (IOError, MetatabError) as e:
                err("Failed to open package '{}' from reference '{}': {}".format(package_url, dist.url, e))

            for r in p.resources():

                mimetype = mimetypes.guess_type(r.resolved_url.path)[0]

                try:
                    ext = mimetypes.guess_extension(mimetype)[1:]
                except:
                    ext = None

                d = dict(
                    name=r.name,
                    format = ext,
                    url=str(r.resolved_url),
                    mimetype=mimetype,
                    description=r.markdown
                )

                resources.append(d)
                prt("Adding {} resource {}".format(d['format'], d['name']))

        elif dist.type == 'fs':
            # Fervently hope that this is a web acessible fs distribution
            from requests import HTTPError
            from rowgenerators.exceptions import DownloadError
            try:
                markdown = metadata_url.doc.markdown
            except (HTTPError, DownloadError):
                pass

        else:
            warn("Unknown distribution type '{}' for '{}'  ".format(dist.type, dist.value))

    try:
        pkg['notes'] = markdown or doc.markdown #doc.find_first_value('Root.Description')
    except (OSError, DownloadError) as e:
        warn(e)

    pkg['resources'] = resources

    c.action.package_update(**pkg)

    pkg = c.action.package_show(name_or_id=pkg['id'])

    ##
    ## Add a term with CKAN info.

    doc['Root'].get_or_new_term('CkanId').value = pkg['id']

    if org_name_slug is None and pkg.get('organization'):
        doc['Root'].get_or_new_term('CkanOrg', (pkg.get('organization') or {}).get('name'))

    groups = doc['Root'].find('Group')
    for g in groups:
        doc.remove_term(g)

    for group in pkg.get('groups', []):
        doc['Root'].new_term('Group', group['name'])

    write_doc(doc, m.mt_file)


def configure_ckan(m):
    """Load groups and organizations, from a file in Metatab format"""
    from ckanapi import RemoteCKAN

    try:
        doc = MetapackDoc(m.mt_file, cache=m.cache)
    except (IOError, MetatabError) as e:
        err("Failed to open metatab '{}': {}".format(m.mt_file, e))

    c = RemoteCKAN(m.ckan_url, apikey=m.api_key)

    groups = { g['name']:g for g in c.action.group_list(all_fields=True) }

    for g in doc['Groups']:

        if g.value not in groups:
            prt('Creating group: ', g.value)
            c.action.group_create(name=g.value,
                                  title=g.get_value('title'),
                                  description=g.get_value('description'),
                                  id=g.get_value('id'),
                                  image_url=g.get_value('image_url'))

    orgs = {o['name']: o for o in c.action.organization_list(all_fields=True)}

    for o in doc['Organizations']:

        if o.value not in orgs:
            prt('Creating organization: ', o.value)
            c.action.organization_create(name=o.value,
                                  title=o.get_value('title'),
                                  description=o.get_value('description'),
                                  id=o.get_value('id'),
                                  image_url=o.get_value('image_url'))

def dump_ckan(m):
    """Create a groups and organization file"""

    doc = MetapackDoc(cache=m.cache)
    doc.new_section('Groups',        'Title Description Id Image_url'.split())
    doc.new_section('Organizations', 'Title Description Id Image_url'.split())

    c = RemoteCKAN(m.ckan_url, apikey=m.api_key)

    for g in c.action.group_list(all_fields=True):
        print(g.keys())

    for o in c.action.organization_list(all_fields=True):
        print(g.keys())



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

""" """


from .core import PackageBuilder
from metapack.util import datetime_now

class ExcelPackageBuilder(PackageBuilder):
    """An Excel File Package"""

    type_code = 'xlsx'
    type_suffix = '.xlsx'

    def __init__(self, source_ref=None, package_root=None,  callback=None, env=None):
        super().__init__(source_ref, package_root, callback, env)

        self.package_path, self.cache_path = self.make_package_path(self.package_root, self.package_name)

        self.cache_path = self.package_name + self.type_suffix

        self.package_path = self.package_root.join(self.cache_path)

        self.package_root.ensure_dir()

    @classmethod
    def make_package_path(cls, package_root, package_name):

        cache_path = package_name + cls.type_suffix

        package_path = package_root.join(cache_path)

        return package_path, cache_path

    def save(self):
        from openpyxl import Workbook
        from openpyxl.cell import WriteOnlyCell
        from openpyxl.styles import PatternFill, Alignment

        self.check_is_ready()

        self.wb = Workbook(write_only=True)

        meta_ws = self.wb.create_sheet()
        meta_ws.title = "meta"
        meta_ws.sheet_properties.tabColor = "8888ff"

        meta_ws.column_dimensions['A'].width = 15
        meta_ws.column_dimensions['B'].width = 40
        meta_ws.column_dimensions['C'].width = 20
        meta_ws.column_dimensions['D'].width = 20
        meta_ws.column_dimensions['E'].width = 20

        self.sections.resources.sort_by_term()

        self.load_declares()

        self.doc.cleanse()

        self._load_resources()

        self._clean_doc()

        fill = PatternFill("solid", fgColor="acc0e0")  # PatternFill(patternType='gray125')
        table_fill = PatternFill("solid", fgColor="d9dce0")  # PatternFill(patternType='gray125')

        alignment = Alignment(wrap_text=False)

        self._doc['Root'].get_or_new_term('Root.Issued').value = datetime_now()

        for i, row in enumerate(self.doc.rows, 1):

            if row[0] == 'Section' or row[0] == 'Table':
                styled_row = []
                for c in row + [''] * 5:
                    cell = WriteOnlyCell(meta_ws, value=c)
                    cell.fill = fill if row[0] == 'Section' else table_fill
                    styled_row.append(cell)
                meta_ws.append(styled_row)


            else:
                meta_ws.append(row)


        self.wb.save(self.package_path.path)

        return self.package_path

    def _load_resources(self):
        """Remove the geography from the files, since it isn't particularly useful in Excel"""

        for t in self.doc.find('Root.Table'):
            for c in t.find('Table.Column'):
                if c.get_value('datatype') == 'geometry':
                    c['transform'] = '^empty_str'
                    c['datatype'] = 'text'

        return super()._load_resources()

    def _load_resource(self, source_r, abs_path=False):

        r = self._doc.resource(source_r.name)

        from itertools import islice

        self.prt("Loading data for sheet '{}' ".format(r.name))

        ws = self.wb.create_sheet(r.name)

        ws.append(r.headers)


        for row in islice(r, 1, None):
            ws.append(row)

        r.url = r.name


# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""Support for IPython and Python kernels in Jupyter Notebooks"""

from metapack import open_package as op
from metapack.constants import PACKAGE_PREFIX
from os import getcwd
from metapack.util import walk_up
from os.path import getmtime, join, exists
from metapack.exc import PackageError
from rowgenerators.exceptions import RowGeneratorError


def caller_locals():
    """Get the local variables in the caller's frame."""
    import inspect
    frame = inspect.currentframe()
    try:
        return frame.f_back.f_back.f_locals
    finally:
        del frame


def in_build():
    """Return True if running in a build, rather than interactively in Jupyter"""
    return 'metatab_doc' in caller_locals()


def open_source_package(dr=None):
    """Like open_package(), but always open the source package"""
    if dr is None:
        dr = getcwd()

    for i, e in enumerate(walk_up(dr)):

        if 'metadata.csv' in e[2]:
            return op(join(e[0], 'metadata.csv'))

        if i > 2:
            break

    return None


def open_package(locals=None, dr=None):
    """Try to open a package with the metatab_doc variable, which is set when a Notebook is run
    as a resource. If that does not exist, try the local _packages directory"""

    if locals is None:
        locals = caller_locals()

    try:
        # Running in a package build
        return op(locals['metatab_doc'])

    except KeyError:
        # Running interactively in Jupyter

        package_name = None
        build_package_dir = None
        source_package = None

        if dr is None:
            dr = getcwd()

        for i, e in enumerate(walk_up(dr)):

            if 'metadata.csv' in e[2]:
                source_package = join(e[0], 'metadata.csv')
                p = op(source_package)
                package_name = p.find_first_value("Root.Name")

                if not package_name:
                    raise PackageError("Source package in {} does not have root.Name term".format(e[0]))

                if PACKAGE_PREFIX in e[1]:
                    build_package_dir = join(e[0], PACKAGE_PREFIX)

                break

            if i > 2:
                break

        if build_package_dir and package_name and exists(join(build_package_dir, package_name)):
            # Open the previously built package
            built_package = join(build_package_dir, package_name)
            try:
                return op(built_package)
            except RowGeneratorError as e:
                pass  # Probably could not open the metadata file.

        if source_package:
            # Open the source package
            return op(source_package)

    raise PackageError("Failed to find package, either in locals() or above dir '{}' ".format(dr))


def rebuild_schema(doc, r, df):
    """Rebuild the schema for a resource based on a dataframe"""
    import numpy as np

    # Re-get the resource in the doc, since it may be different.
    try:
        r = doc.resource(r.name)
    except AttributeError:
        # Maybe r is actually a resource name
        r = doc.resource(r)

    def alt_col_name(name, i):
        import re

        if not name:
            return 'col{}'.format(i)

        return re.sub('_+', '_', re.sub('[^\w_]', '_', str(name)).lower()).rstrip('_')

    df_types = {
        np.dtype('O'): 'text',
        np.dtype('int64'): 'integer',
        np.dtype('float64'): 'number'
    }

    try:
        df_index_frame = df.index.to_frame()
    except AttributeError:
        df_index_frame = None

    def get_col_dtype(c):

        c = str(c)

        try:
            return df_types[df[c].dtype]
        except KeyError:
            # Maybe it is in the index?
            pass

        try:
            return df_types[df_index_frame[c].dtype]
        except TypeError:
            # Maybe not a multi-index
            pass

        if c == 'id' or c == df.index.name:
            return df_types[df.index.dtype]

        return 'unknown'

    columns = []
    schema_term = r.schema_term[0]

    if schema_term:

        old_cols = {c['name'].value: c.properties for c in schema_term.children}
        for c in schema_term.children:
            schema_term.remove_child(c)

        schema_term.children = []

    else:
        old_cols = {}
        schema_term = doc['Schema'].new_term('Table', r.schema_name)

    index_names = [n if n else "id" for n in df.index.names]

    for i, col in enumerate(index_names + list(df.columns)):
        acn = alt_col_name(col, i) if alt_col_name(col, i) != str(col) else ''

        d = {'name': col, 'datatype': get_col_dtype(col), 'altname': acn}

        if col in old_cols.keys():
            lookup_name = col
        elif acn in old_cols.keys():
            lookup_name = acn
        else:
            lookup_name = None

        if lookup_name and lookup_name in old_cols:

            for k, v in schema_term.properties.items():

                old_col = old_cols.get(lookup_name)

                for k, v in old_col.items():
                    if k != 'name' and v:
                        d[k] = v

        columns.append(d)

    for c in columns:
        name = c['name']
        del c['name']
        datatype = c['datatype']
        del c['datatype']
        altname = c['altname']
        del c['altname']

        schema_term.new_child('Column', name, datatype=datatype, altname=altname, **c)


def rewrite_schema(r, df, doc=None):
    """Rebuild the schema for a resource based on a dataframe and re-write the doc"""

    from metapack.cli.core import write_doc

    if doc is None:
        doc = open_source_package()

    rebuild_schema(doc, r, df)

    write_doc(doc, doc.ref)


def interactive_rewrite_schema(r, df, doc=None):
    """Rebuild the schema for a resource based on a dataframe and re-write the doc,
    but only if running the notebook interactively, not while building"""

    if 'metatab_doc' in caller_locals():
        return False

    if doc is None:
        doc = open_source_package()

    rewrite_schema(r, df, doc)

    return True


def get_dataframes():
    """Yield tuples of dataframe variable name and the dataframe. Skips variables with names that
    start with an underscore"""

    for k, v in caller_locals().items():
        if k.startswith('_'):
            continue

        if isinstance(v, pd.core.frame.DataFrame):
            yield k, v


import json
import os.path
import re
import ipykernel
import requests
from urllib.parse import urljoin
from notebook.notebookapp import list_running_servers

def get_notebook_path():
    """
    Return the full path of the jupyter notebook.
    """
    kernel_id = re.search('kernel-(.*).json',
                          ipykernel.connect.get_connection_file()).group(1)
    servers = list_running_servers()
    for ss in servers:
        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                params={'token': ss.get('token', '')})
        for nn in json.loads(response.text):
            if nn['kernel']['id'] == kernel_id:
                relative_path = nn['notebook']['path']
                return os.path.join(ss['notebook_dir'], relative_path)


def get_notebook_rel_path(pkg=None):
    """Get the path of a notebook, relative to the current soruce package"""
    pkg = pkg or open_source_package()
    pkg_path = str(pkg.package_url.fspath)
    nb_path = get_notebook_path()

    return nb_path.replace(pkg_path, '').strip('/')


def add_dataframe(df, name, pkg=None, description=''):
    """Add a dataframe to a source package. Pass in either the name of the dataframe, or the dataframe.
    If the dataframeis passed it, the name will be the dataframe's variable name. The function
    will re-write the source package with the new resource.

    """
    from warnings import warn
    from metapack.cli.core import alt_col_name, type_map
    import numpy as np

    if name is None or df is None:
        warn("Did not find dataframe for reference '{}' ".format(ref))
        return

    pkg = pkg or open_source_package()

    resource_ref = 'file:' + get_notebook_rel_path(pkg) + '#' + name

    t = pkg.find_first('Root.Datafile', value=resource_ref)
    col_props = {}

    if t:
        print("Datafile exists for url '{}', deleting".format(resource_ref))

        if t.schema_term:
            col_props = { c['name']:c for c in t.columns()}
            pkg.remove_term(t.schema_term)

        pkg.remove_term(t)

    t = pkg['Resources'].new_term('Root.Datafile', resource_ref, name=name, description=description)

    st = pkg['Schema'].new_term('Table', t.schema_name, description=description)

    for i, name in enumerate(df.columns):

        props = col_props.get(name,{})

        try:
            native_type = type(np.asscalar(df[name].dtype.type(0))).__name__
        except ValueError:
            native_type = df[name].dtype.name
        except AttributeError:
            native_type = type(df[name][0]).__name__

        for pn in 'datatype name pos header'.split():
            if pn in props:
                del props[pn]

        if 'altname' in props:
            altname = props['altname']
            del props['altname']
        else:
            raw_alt_name = alt_col_name(name, i)
            altname = raw_alt_name if raw_alt_name != name else ''

        col = df[name]

        if hasattr(col, 'description'):  # custom property
            props['description'] = col.description

        t = st.new_child('Column', name,
                            datatype=type_map.get(native_type, native_type),
                            altname=altname,
                            **props)


    pkg.write_csv()


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
Tests for the CLI
"""

from pkg_resources import iter_entry_points, get_distribution, DistributionNotFound
from metapack.cli.core import prt, err
from shlex import split
import argparse
import sys

# from https://stackoverflow.com/a/10743550
import contextlib
@contextlib.contextmanager
def capture():
    import sys
    from io import StringIO
    oldout,olderr = sys.stdout, sys.stderr
    try:
        out=[StringIO(), StringIO()]
        sys.stdout,sys.stderr = out
        yield out
    finally:
        sys.stdout,sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()

def exec_cmd(args):
    """"Execute a CLI command, through the backdoor"""

    parser = argparse.ArgumentParser(
        prog='mp',
        description='Create and manipulate metatab data packages. ')

    subparsers = parser.add_subparsers(help='Commands')

    for ep in iter_entry_points(group='mt.subcommands'):
        f = ep.load()
        f(subparsers)

    if isinstance(args, str):
        args = split(args)

    args = parser.parse_args(args)

    try:
        with capture() as c:
            args.run_command(args)
        exc = None
    except Exception as e:
        exc = e
    except SystemExit as e:
        exc = e


    return c+[exc]


def delete_contents(directory):
    import os
    import glob
    import shutil

    files = glob.glob(directory+'/*')

    for f in files:
        shutil.rmtree(f)


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Common support for jupyter notebooks.
"""
import nbformat
from os.path import join, dirname
from os import makedirs
import logging

import unicodecsv as csv

logger = logging.getLogger('user')
err_logger = logging.getLogger('cli-errors')
debug_logger = logging.getLogger('debug')
doc_logger = logging.getLogger('doc')

log_init = False

def init_logging(log_level=logging.INFO):
    import sys

    global log_init

    if log_init:
        return

    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter('%(message)s'))
    out_hdlr.setLevel(log_level)
    logger.addHandler(out_hdlr)
    logger.setLevel(log_level)

    out_hdlr = logging.StreamHandler(sys.stderr)
    out_hdlr.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    out_hdlr.setLevel(logging.WARN)
    err_logger.addHandler(out_hdlr)
    err_logger.setLevel(logging.WARN)

    out_hdlr = logging.StreamHandler(sys.stderr)
    out_hdlr.setFormatter(logging.Formatter('DBG %(levelname)s: %(message)s'))
    out_hdlr.setLevel(logging.INFO)
    debug_logger.addHandler(out_hdlr)
    debug_logger.setLevel(logging.CRITICAL)

    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter('%(message)s'))
    out_hdlr.setLevel(log_level)
    doc_logger.addHandler(out_hdlr)
    doc_logger.setLevel(log_level)

    log_init = True


def ensure_source_package_dir(nb_path, pkg_name):
    """Ensure all of the important directories in a source package exist"""

    pkg_path = join(dirname(nb_path), pkg_name)

    makedirs(join(pkg_path,'notebooks'),exist_ok=True)
    makedirs(join(pkg_path, 'docs'), exist_ok=True)

    return pkg_path


def get_metatab_doc(nb_path):
    """Read a notebook and extract the metatab document. Only returns the first document"""

    from metatab.generate import TextRowGenerator, CsvDataRowGenerator
    from metatab import MetatabDoc

    with open(nb_path) as f:
        nb = nbformat.reads(f.read(), as_version=4)

    for cell in nb.cells:
        source = ''.join(cell['source']).strip()
        if source.startswith('%%metatab'):
            return MetatabDoc(TextRowGenerator(source))


def get_package_dir(nb_path):
    """Return the package directory for a Notebook that has an embeded Metatab doc, *not* for
    notebooks that are part of a package """
    doc = get_metatab_doc(nb_path)
    doc.update_name(force=True, create_term=True)
    pkg_name = doc['Root'].get_value('Root.Name')
    assert pkg_name

    return ensure_source_package_dir(nb_path, pkg_name), pkg_name


def process_schema(doc, resource, df):
    """Add schema entiries to a metatab doc from a dataframe"""
    from rowgenerators import SourceError
    from requests.exceptions import ConnectionError

    from metapack.cli.core import extract_path_name, alt_col_name, type_map
    from tableintuit import TypeIntuiter
    from metapack.rowgenerator import PandasDataframeSource
    from appurl import parse_app_url

    try:
        doc['Schema']
    except KeyError:
        doc.new_section('Schema', ['DataType', 'Altname', 'Description'])

    schema_name = resource.get_value('schema', resource.get_value('name'))

    schema_term = doc.find_first(term='Table', value=schema_name, section='Schema')

    if schema_term:
        logger.info("Found table for '{}'; skipping".format(schema_name))
        return

    path, name = extract_path_name(resource.url)

    logger.info("Processing {}".format(resource.url))

    si = PandasDataframeSource(parse_app_url(resource.url), df, cache=doc._cache, )

    try:
        ti = TypeIntuiter().run(si)
    except SourceError as e:
        logger.warn("Failed to process '{}'; {}".format(path, e))
        return
    except ConnectionError as e:
        logger.warn("Failed to download '{}'; {}".format(path, e))
        return

    table = doc['Schema'].new_term('Table', schema_name)

    logger.info("Adding table '{}' to metatab schema".format(schema_name))

    for i, c in enumerate(ti.to_rows()):
        raw_alt_name = alt_col_name(c['header'], i)
        alt_name = raw_alt_name if raw_alt_name != c['header'] else ''

        t = table.new_child('Column', c['header'],
                            datatype=type_map.get(c['resolved_type'], c['resolved_type']),
                            altname=alt_name,
                            description=df[c['header']].description \
                                        if hasattr(df, 'description') and df[c['header']].description else ''
                            )

    return table


def write_geojson(path_or_flo, columns, gen):
    import fiona
    from fiona.crs import from_epsg
    from collections import OrderedDict

    type_map = {
        'number': 'float'
    }

    schema = {
        'geometry': 'Unknown',
        'properties': OrderedDict(
            [(c['header'], type_map.get(c['datatype'], c['datatype'])) for c in columns]
        )
    }

    try:
        f = fiona.open(path_or_flo,
                       driver='GeoJSON',
                       crs=from_epsg(4326),
                       schema=schema)
    except TypeError:
        f = path_or_flo  # Assume that it's already a file-like-object

    try:
        w = csv.writer(f)
        w.writerow(headers)
        w.writerows(gen)

        try:
            return f.getvalue()
        except AttributeError:
            return None

    finally:
        f.close()


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

import sys
from metapack import Downloader
from metapack.cli.core import prt, err

from rowgenerators import parse_app_url
from rowgenerators.exceptions import AppUrlError
from metapack.index import SearchIndex, search_index_file

downloader = Downloader()

def search(subparsers):

    parser = subparsers.add_parser(
        'search',
        help='Search for packages',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    parser.add_argument('-l', '--list', default=False, action='store_true',
                        help="List the packages that would be indexed ( Only from the JSON index")

    parser.add_argument('-f', '--format', help='Select a specific package format')

    parser.add_argument('-1', '--one', action='store_true',
                        help='Find only one result, using the same resolution process used when building packages')

    parser.add_argument('-c', '--config', default=False, action='store_true',
                        help="Show the path to the index file")

    parser.set_defaults(run_command=run_search)

    parser.add_argument('search', nargs='?', help="Path or URL to a metatab file")

def run_search(args):

    from tabulate import tabulate

    if args.config:
        prt(search_index_file())
        sys.exit(0)

    if not args.search or args.list:

        idx = SearchIndex(search_index_file())

        packages = [ (e['name'], e['format'], e['url']) for e in idx.list()
                     if args.format is None or args.format == e['format']]

        print(tabulate(packages, headers='Name Format Url'.split()))

    elif args.one:

        if args.search.startswith('index'):
            url_s = args.search
        else:
            url_s = 'index:' + args.search

        try:
            u = parse_app_url(url_s)

            prt(str(u.get_resource()))

        except AppUrlError as e:
            err(f"Failed to resolve: {str(u)}; {str(e)}")

    else:

        idx = SearchIndex(search_index_file())

        prt('Index file:', idx.path)

        p = idx.search(args.search, args.format)

        packages = []
        for e in p:
            packages.append((e['name'], e['format'], e['url']))

        print(tabulate(packages, headers='Name Format Url'.split()))







#
#
# Code storage for the dump_resource features of the metapack program

from metapack import MetapackDoc, Downloader

from metapack.cli.core import err, dump_resource, dump_resources


def metatab_query_handler(m):
    if m.args.resource or m.args.head:

        limit = 20 if m.args.head else None

        try:
            doc = MetapackDoc(m.mt_file, cache=m.cache)

        except OSError as e:
            err("Failed to open Metatab doc: {}".format(e))
            return

        if m.resource:
            dump_resource(doc, m.resource, limit)
        else:
            dump_resources(doc)


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Functions for converting Jupyter notebooks
"""
from os import getcwd
from os.path import abspath, normpath, exists, dirname

import nbformat
from nbconvert.writers import FilesWriter
from traitlets.config import Config

from rowgenerators import parse_app_url
from rowgenerators.util import fs_join as join
from metatab import DEFAULT_METATAB_FILE
from metapack import MetapackDoc
from metapack.cli.core import prt, err
from metapack.util import ensure_dir, copytree
from metapack.jupyter.core import logger
from metapack.jupyter.exporters import NotebookExecutor, DocumentationExporter, HugoExporter, WordpressExporter
from metapack.jupyter.preprocessors import ExtractInlineMetatabDoc, ExtractMetatabTerms



def convert_documentation(nb_path):
    """Run only the document conversion portion of the notebook conversion

      The final document will not be completel
    """

    with open(nb_path) as f:
        nb = nbformat.reads(f.read(), as_version=4)

    doc = ExtractInlineMetatabDoc(package_url="metapack+file:" +dirname(nb_path)).run(nb)

    package_name = doc.as_version(None)

    output_dir = join(getcwd(), package_name)

    de = DocumentationExporter(config=Config(), log=logger, metadata=doc_metadata(doc))
    prt('Converting documentation')
    output, resources = de.from_filename(nb_path)

    fw = FilesWriter()

    fw.build_directory = join(output_dir,'docs')
    fw.write(output, resources, notebook_name='notebook')
    prt("Wrote documentation to {}".format(fw.build_directory))


def convert_notebook(nb_path):

    prt('Convert notebook to Metatab source package')

    if not exists(nb_path):
        err("Notebook path does not exist: '{}' ".format(nb_path))

    c = Config()

    pe = NotebookExecutor(config=c, log=logger)

    prt('Running the notebook')
    output, resources = pe.from_filename(nb_path)

    fw = FilesWriter()
    fw.build_directory = pe.output_dir

    fw.write(output, resources, notebook_name=DEFAULT_METATAB_FILE)

    de = DocumentationExporter(config=c, log=logger, metadata=doc_metadata(pe.doc))

    prt('Exporting documentation')
    output, resources = de.from_filename(nb_path)

    fw.build_directory = join(pe.output_dir,'docs')
    fw.write(output, resources, notebook_name='notebook')

    new_mt_file = join(pe.output_dir, DEFAULT_METATAB_FILE)

    doc = MetapackDoc(new_mt_file)

    de.update_metatab(doc, resources)

    for lib_dir in pe.lib_dirs:

        lib_dir = normpath(lib_dir).lstrip('./')

        doc['Resources'].new_term("Root.PythonLib", lib_dir)

        path = abspath(lib_dir)
        dest = join(pe.output_dir, lib_dir)

        ensure_dir(dest)
        copytree(path, join(pe.output_dir, lib_dir))

    doc.write_csv()

    # Reset the input to use the new data

    prt('Running with new package file: {}'.format(new_mt_file))


def extract_metatab(nb_path):

    if not exists(nb_path):
        err("Notebook path does not exist: '{}' ".format(nb_path))

    c = Config()

    with open(nb_path) as f:
        nb = nbformat.reads(f.read(), as_version=4)

    return ExtractInlineMetatabDoc(package_url="metapack+file:" + dirname(nb_path)).run(nb)


def doc_metadata(doc):
    """Create a metadata dict from a MetatabDoc, for Document conversion"""

    r = doc['Root'].as_dict()
    r.update(doc['Contacts'].as_dict())
    r['author'] = r.get('author', r.get('creator', r.get('wrangler')))

    return r

def convert_hugo(nb_path, hugo_path):
    from os import environ

    # Total hack. Would like the -H to be allowed to have no arg, and then use the env var,
    # but I don't know how to do that. This is the case where the user types
    # -H nb_path, so just go with it.
    if hugo_path and not nb_path:
        nb_path = hugo_path
        hugo_path = environ.get('METAPACK_HUGO_DIR')

    if not hugo_path:
        err("Must specify value for -H or the METAPACK_HUGO_DIR environment var")

    if not exists(nb_path):
        err("Notebook path does not exist: '{}' ".format(nb_path))

    c = Config()
    c.HugoExporter.hugo_dir = hugo_path
    he = HugoExporter(config=c,log=logger)

    output, resources = he.from_filename(nb_path)

    prt('Writing Notebook to Hugo Markdown')

    prt('    Writing ', resources['unique_key']+resources['output_extension'])
    for k, v in resources['outputs'].items():
        prt('    Writing ',k)

    fw = FilesWriter()
    fw.write(output, resources, notebook_name=resources['unique_key'])

def convert_wordpress(nb_path, hugo_path):
    from os import environ

    if not exists(nb_path):
        err("Notebook path does not exist: '{}' ".format(nb_path))

    c = Config()
    c.WordpressExporter.staging_dir = hugo_path
    he = WordpressExporter(config=c, log=logger)

    output, resources = he.from_filename(nb_path)

    prt('Writing Notebook to Wordpress HTML')

    output_file = resources['unique_key'] + resources['output_extension']
    prt('    Writing ', output_file)

    resource_outputs = []

    for k, v in resources['outputs'].items():
        prt('    Writing ', k)
        resource_outputs.append(k)

    fw = FilesWriter()
    fw.write(output, resources, notebook_name=resources['unique_key'])

    return output_file, resource_outputs






# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""Support for IPython and Python kernels in Jupyter Notebooks"""


from .metapack import MetapackExporter
from .exporters import HugoExporter
from .ipython import open_package, open_source_package

def init():
    """Initialize features that are normally initialized in the CLI"""

    from metapack.appurl import SearchUrl
    import metapack as mp
    from os import environ

    SearchUrl.initialize()  # This makes the 'index:" urls work
    mp.Downloader.context.update(environ)  # Allows resolution of environmental variables in urls


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
CLI program for managing packages
"""

import re

from metapack import MetapackDoc, Downloader
from metapack.cli.core import prt, warn, write_doc, \
    update_name, extract_path_name, MetapackCliMemo
from rowgenerators import parse_app_url
from rowgenerators.exceptions import SourceError, RowGeneratorError
from tableintuit import RowIntuitError

downloader = Downloader()


def url(subparsers):
    parser = subparsers.add_parser(
        'url',
        help='Add resource urls to a package',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    parser.set_defaults(run_command=run_url)

    parser.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    parser.set_defaults(handler=None)

    group = parser.add_mutually_exclusive_group()

    group.add_argument('-a', '--add', default=False,
                       help='Add a file or url to the resources. With a directory add a data files in the directory. '
                            'If given a URL to a web page, will add all links that point to CSV, Excel Files and '
                            'data files in ZIP files. (Caution: it will download and cache all of these files. )')

    group.add_argument('-E', '--enumerate',
                       help='Enumerate the resources referenced from a URL. Does not alter the Metatab file')


def run_url(args):
    m = MetapackCliMemo(args, downloader)

    if m.args.add:
        update_name(m.mt_file, fail_on_missing=False, report_unchanged=False)

        add_resource(m.mt_file, m.args.add, cache=m.cache)

    if m.args.enumerate:

        from metatab.util import enumerate_contents

        specs = list(enumerate_contents(m.args.enumerate, m.cache, callback=prt))

        for s in specs:
            prt(classify_url(s.url), s.target_format, s.url, s.target_segment)


def classify_url(url):
    ss = parse_app_url(url)

    if ss.target_format in DATA_FORMATS:
        term_name = 'DataFile'
    elif ss.target_format in DOC_FORMATS:
        term_name = 'Documentation'
    else:
        term_name = 'Datafile'

    return term_name


def add_resource(mt_file, ref, cache):
    """Add a resources entry, downloading the intuiting the file, replacing entries with
    the same reference"""

    if isinstance(mt_file, MetapackDoc):
        doc = mt_file
    else:
        doc = MetapackDoc(mt_file)

    if not 'Resources' in doc:
        doc.new_section('Resources')

    doc['Resources'].args = [e for e in set(doc['Resources'].args + ['Name', 'StartLine', 'HeaderLines', 'Encoding']) if
                             e]

    seen_names = set()

    u = parse_app_url(ref)

    # The web and file URLs don't list the same.

    if u.proto == 'file':
        entries = u.list()
    else:
        entries = [ssu for su in u.list() for ssu in su.list()]

    errors = []

    for e in entries:
        if not add_single_resource(doc, e, cache=cache, seen_names=seen_names):
            errors.append(e)


    if errors:
        prt()
        warn("Found, but failed to add these urls:")
        for e in errors:
            print('    ', e)

    write_doc(doc, mt_file)


def add_single_resource(doc, ref, cache, seen_names):
    from metatab.util import slugify

    t = doc.find_first('Root.Datafile', value=ref)

    if t:
        prt("Datafile exists for '{}', deleting".format(ref))
        doc.remove_term(t)

    term_name = classify_url(ref)

    path, name = extract_path_name(ref)

    # If the name already exists, try to create a new one.
    # 20 attempts ought to be enough.
    if name in seen_names:
        base_name = re.sub(r'-?\d+$', '', name)

        for i in range(1, 20):
            name = "{}-{}".format(base_name, i)
            if name not in seen_names:
                break

    seen_names.add(name)

    encoding = start_line = None
    header_lines = []

    try:
        encoding, ri = run_row_intuit(path, cache)

        start_line = ri.start_line or None
        header_lines = ri.header_lines
    except RowIntuitError as e:
        warn("Failed to intuit '{}'; {}".format(ref, e))
    except RowGeneratorError as e:
        warn("Can't generate rows for: '{}'; {}".format(ref, e))
        return None
    except SourceError as e:
        warn("Source Error: '{}'; {}".format(ref, e))
        return None

    except Exception as e:
        warn("Error: '{}'; {}".format(ref, e))
        raise

    if not name:
        from hashlib import sha1
        name = sha1(slugify(path).encode('ascii')).hexdigest()[:12]

        # xlrd gets grouchy if the name doesn't start with a char
        try:
            int(name[0])
            name = 'a' + name[1:]
        except:
            pass

    prt("Added {}, url: {} ".format(name, ref))

    return doc['Resources'].new_term(term_name, ref, name=name,
                                     startline=start_line,
                                     headerlines=','.join(str(e) for e in header_lines)
                                     )

def run_row_intuit(path, cache):
    from tableintuit import RowIntuiter
    from itertools import islice
    from rowgenerators.exceptions import TextEncodingError

    for encoding in ('ascii', 'utf8', 'latin1'):
        try:

            u = parse_app_url(path)
            u.encoding = encoding

            rows = list(islice(u.get_resource().get_target().generator, 5000))
            ri = RowIntuiter().run(list(rows))
            return encoding, ri
        except (TextEncodingError, UnicodeEncodeError, UnicodeDecodeError) as e:
            pass # Try the next encoding


    raise RowIntuitError('Failed to convert with any encoding')


DATA_FORMATS = ('xls', 'xlsx', 'tsv', 'csv')
DOC_FORMATS = ('pdf', 'doc', 'docx', 'html')



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""

import json
from os import rename
from shutil import copy
from os.path import exists
from os import rename


def search_index_file():
    from metapack import Downloader
    return Downloader().cache.getsyspath('index.json')

class SearchIndex(object):

    pkg_format_priority = {
        'fs': 5,
        'zip': 4,
        'xlsx': 3,
        'csv': 2,
        'source': 1
    }

    def __init__(self, path):
        self.path = path

        self._db = None

    def open(self):
        if not self._db:
            try:
                with open(self.path) as f:
                    self._db = json.load(f)
            except FileNotFoundError:
                self._db = {}

    def clear(self):

        self._db = {}
        self.write()

    def write(self):
        """Safely write the index data to the index file """
        index_file = self.path
        new_index_file = index_file + '.new'
        bak_index_file = index_file + '.bak'

        with open(new_index_file, 'w') as f:
            json.dump(self._db, f, indent=4)

        if exists(index_file):
            copy(index_file, bak_index_file)

        rename(new_index_file, index_file)

    def _make_package_entry(self, ident, name, nvname, version, format, url):
        """

        """

        self.open()

        try:
            version = 'V'+str(int(version)).zfill(10)
        except ValueError:
            pass

        if format is None:
            format = 'fs'

        self._db[ident] = {'t':'ident', 'ref':nvname } # these should always be equivalent

        self._db[name] = {'t':'name',  'ref': nvname, 'version': version , 'ident': ident}

        if not nvname in self._db:
            self._db[nvname] = {
                't': 'nvname',
                'packages': {}
            }

        key = '{}-{}'.format(name, format)

        self._db[nvname]['packages'][key] = {
            'name': name,
            'nvname': nvname,
            'version': version,
            'format': format,
            'ident': ident,
            'url': url
        }

    def add_package(self, pkg):
        from os.path import abspath

        from metapack.package import open_package
        ref_url = pkg.package_url.clone()
        ref_url.path = abspath(ref_url.path)

        ref = str(ref_url)

        nv_name = pkg._generate_identity_name(mod_version=None)  # Non versioned-name, for the latest package
        name = pkg.name
        version = pkg.get_value('Root.Version')
        identifier = pkg.get_value('Root.Identifier')

        if not pkg.get_value('Root.Issued'):
            format = 'source'
        else:
            format = ref_url.get_resource().get_target().target_format

        self._make_package_entry(identifier, name, nv_name, version, format, ref)

    def add_entry(self, ident, name, nvname, version, format, url):
        self._make_package_entry(ident, name, nvname, version, format, url)

    def update(self,o):
        """Update from another index or index dict"""

        self.open()

        try:
            self._db.update(o._db)
        except AttributeError:
            self._db.update(o)


    def list(self):

        packages = []

        self.open()

        for k, v in  self._db.items():
            if v.get('t') == 'nvname':
                for pkey, e in v.get('packages').items():
                    packages.append(e)

        return list(reversed(sorted(packages, key=lambda x: (x['name'], x['version'], self.pkg_format_priority[x['format']]))))

    def records(self):
        self.open()
        for k,v in self._db.items():
            if v['t'] == 'nvname':
                for pk, pv in v['packages'].items():
                    yield [pv['name'], pv['nvname'], pv['version'], pv['format'], pv['url']]

    def search(self, key, format='issued'):
        from rowgenerators import parse_app_url

        url = parse_app_url(key)

        search_term = url.path

        self.open()

        if format == 'all':
            format = None
        elif format == 'issued':
            format = ['zip','csv','xlsx','fs']

        if format and not isinstance(format, (list, tuple)):
            format = [format]

        record = self._db.get(search_term) # Case when the term is a name, nvname, or ident

        if record:
            records = [record]
            match_type = 'exact'
        else:
            records = [ record for k, record in self._db.items() if search_term in k]
            match_type = 'subset'

        packages = []
        seen = set()
        for e in records:

            key_type = e['t']

            if key_type == 'ident':
                match_key = 'nvname'
                match_value = e['ref']
            else:
                match_key = key_type
                match_value = search_term

            if e.get('ref'):
                e = self._db[e['ref']]

            for pkey, p in e['packages'].items():

                if (match_type == 'exact' and p[match_key] == match_value) or \
                   (match_type == 'subset' and search_term in p[match_key]):
                    if (format and p['format'] in format) or format is None:
                        if not p['name'] in seen:
                            packages.append(p)
                            seen.add(p['name'])

        return list(reversed(sorted(packages, key=lambda x: (x['version'], self.pkg_format_priority[x['format']]) )))


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for managing Metatab files
"""


import json
import sys

from metatab import DEFAULT_METATAB_FILE, resolve_package_metadata_url, MetatabDoc
from metatab.cli.core import prt, new_metatab_file, err, dump_resource, dump_resources, dump_schema, cli_init
from appurl import Url, get_cache
from rowgenerators.util import clean_cache


def metatab():
    import argparse
    parser = argparse.ArgumentParser(
        prog='metatab',
        description='Matatab file parser')

    parser.add_argument('-C', '--clean-cache', default=False, action='store_true',
                        help="Clean the download cache")

    g = parser.add_mutually_exclusive_group(required=True)

    g.add_argument('-i', '--info', default=False, action='store_true',
                   help="Show configuration information")

    g.add_argument('-c', '--create', action='store', nargs='?', default=False,
                   help="Create a new metatab file, from named template. With no argument, uses the 'metatab' template ")

    g.add_argument('-t', '--terms', default=False, action='store_true',
                   help='Parse a file and print out the stream of terms, before interpretation')

    g.add_argument('-I', '--interp', default=False, action='store_true',
                   help='Parse a file and print out the stream of terms, after interpretation')

    g.add_argument('-j', '--json', default=False, action='store_true',
                   help='Parse a file and print out a JSON representation')

    g.add_argument('-y', '--yaml', default=False, action='store_true',
                   help='Parse a file and print out a YAML representation')

    g.add_argument('-R', '--resource', default=False, action='store_true',
                   help='If the URL has no fragment, dump the resources listed in the metatab file. With a fragment, dump a resource as a CSV')

    g.add_argument('-H', '--head', default=False, action='store_true',
                   help="Dump the first 20 lines of a resoruce ")

    g.add_argument('-S', '--schema',
                   help='Dump the schema for one named resource')

    parser.add_argument('-d', '--show-declaration', default=False, action='store_true',
                        help='Parse a declaration file and print out declaration dict. Use -j or -y for the format')

    parser.add_argument('-D', '--declare', help='Parse and incorporate a declaration before parsing the file.' +
                                                ' (Adds the declaration to the start of the file as the first term. )')

    parser.add_argument('file', nargs='?', default=DEFAULT_METATAB_FILE, help='Path to a Metatab file')

    args = parser.parse_args(sys.argv[1:])

    # Specing a fragment screws up setting the default metadata file name
    if args.file.startswith('#'):
        args.file = DEFAULT_METATAB_FILE + args.file

    cache = get_cache('metapack')


    if args.clean_cache:
        clean_cache(cache)

    if args.create is not False:
        new_metatab_file(args.file, args.create)
        exit(0)

    if args.resource or args.head:

        limit = 20 if args.head else None

        u = Url(args.file)
        resource = u.parts.fragment
        metadata_url = u.rebuild_url(False, False)

        package_url, metadata_url = resolve_package_metadata_url(metadata_url)

        try:
            doc = MetatabDoc(metadata_url, cache=cache)
        except OSError as e:
            err("Failed to open Metatab doc: {}".format(e))
            return # Never reached

        if resource:
            dump_resource(doc, resource, limit)
        else:
            dump_resources(doc)


        exit(0)

    if args.show_declaration:

        doc = MetatabDoc()
        doc.load_declarations([args.file])

        print(json.dumps({
            'terms': doc.decl_terms,
            'sections': doc.decl_sections
        }, indent=4))
        exit(0)
    else:

        package_url, metadata_url = resolve_package_metadata_url(args.file)
        try:
            doc = MetatabDoc(metadata_url, cache=cache)
        except IOError as e:
            raise
            err("Failed to open '{}': {}".format(metadata_url, e))

    if args.terms:
        for t in doc._term_parser:
            print(t)

    elif args.json:
        print(json.dumps(doc.as_dict(), indent=4))


    elif args.yaml:
        import yaml
        print(yaml.safe_dump(doc.as_dict(), default_flow_style=False, indent=4))


    elif args.schema:
        dump_schema(doc, args.schema)

    exit(0)